from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.events import AssetKey, AssetObservation, CoercibleToAssetKey
from dagster._core.definitions.metadata import (
    ArbitraryMetadataMapping,
    MetadataValue,
)
from dagster._core.definitions.partition import PartitionsSubset
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.time_window_partitions import TimeWindow, TimeWindowPartitionsSubset
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.instance import DagsterInstance, DynamicPartitionsStore

if TYPE_CHECKING:
    from dagster._core.definitions import PartitionsDefinition
    from dagster._core.definitions.op_definition import OpDefinition
    from dagster._core.definitions.resource_definition import Resources
    from dagster._core.events import DagsterEvent
    from dagster._core.execution.context.system import StepExecutionContext
    from dagster._core.log_manager import DagsterLogManager
    from dagster._core.types.dagster_type import DagsterType

    from .output import OutputContext


class InputContext:
    """The ``context`` object available to the load_input method of :py:class:`RootInputManager`.

    Users should not instantiate this object directly. In order to construct
    an `InputContext` for testing an IO Manager's `load_input` method, use
    :py:func:`dagster.build_input_context`.

    Example:
    .. code-block:: python

        from dagster import IOManager, InputContext

        class MyIOManager(IOManager):
            def load_input(self, context: InputContext):
                ...

    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        job_name: Optional[str] = None,
        op_def: Optional["OpDefinition"] = None,
        config: Optional[Any] = None,
        metadata: Optional[ArbitraryMetadataMapping] = None,
        upstream_output: Optional["OutputContext"] = None,
        dagster_type: Optional["DagsterType"] = None,
        log_manager: Optional["DagsterLogManager"] = None,
        resource_config: Optional[Mapping[str, Any]] = None,
        resources: Optional[Union["Resources", Mapping[str, Any]]] = None,
        step_context: Optional["StepExecutionContext"] = None,
        asset_key: Optional[AssetKey] = None,
        partition_key: Optional[str] = None,
        asset_partitions_subset: Optional[PartitionsSubset] = None,
        asset_partitions_def: Optional["PartitionsDefinition"] = None,
        instance: Optional[DagsterInstance] = None,
    ):
        from dagster._core.definitions.resource_definition import IContainsGenerator, Resources
        from dagster._core.execution.build_resources import build_resources

        self._name = name
        self._job_name = job_name
        self._op_def = op_def
        self._config = config
        self._metadata = metadata or {}
        self._upstream_output = upstream_output
        self._dagster_type = dagster_type
        self._log = log_manager
        self._resource_config = resource_config
        self._step_context = step_context
        self._asset_key = asset_key
        if self._step_context and self._step_context.has_partition_key:
            self._partition_key: Optional[str] = self._step_context.partition_key
        else:
            self._partition_key = partition_key

        self._asset_partitions_subset = asset_partitions_subset
        self._asset_partitions_def = asset_partitions_def

        if isinstance(resources, Resources):
            self._resources_cm = None
            self._resources = resources
        else:
            self._resources_cm = build_resources(
                check.opt_mapping_param(resources, "resources", key_type=str)
            )
            self._resources = self._resources_cm.__enter__()
            self._resources_contain_cm = isinstance(self._resources, IContainsGenerator)
            self._cm_scope_entered = False

        self._events: List["DagsterEvent"] = []
        self._observations: List[AssetObservation] = []
        self._instance = instance

    def __enter__(self):
        if self._resources_cm:
            self._cm_scope_entered = True
        return self

    def __exit__(self, *exc):
        if self._resources_cm:
            self._resources_cm.__exit__(*exc)

    def __del__(self):
        if self._resources_cm and self._resources_contain_cm and not self._cm_scope_entered:
            self._resources_cm.__exit__(None, None, None)

    @property
    def instance(self) -> DagsterInstance:
        if self._instance is None:
            raise DagsterInvariantViolationError(
                "Attempting to access instance, "
                "but it was not provided when constructing the InputContext"
            )
        return self._instance

    @public
    @property
    def has_input_name(self) -> bool:
        """If we're the InputContext is being used to load the result of a run from outside the run,
        then it won't have an input name.
        """
        return self._name is not None

    @public
    @property
    def name(self) -> str:
        """The name of the input that we're loading."""
        if self._name is None:
            raise DagsterInvariantViolationError(
                "Attempting to access name, "
                "but it was not provided when constructing the InputContext"
            )

        return self._name

    @property
    def job_name(self) -> str:
        if self._job_name is None:
            raise DagsterInvariantViolationError(
                "Attempting to access job_name, "
                "but it was not provided when constructing the InputContext"
            )
        return self._job_name

    @public
    @property
    def op_def(self) -> "OpDefinition":
        """The definition of the op that's loading the input."""
        if self._op_def is None:
            raise DagsterInvariantViolationError(
                "Attempting to access op_def, "
                "but it was not provided when constructing the InputContext"
            )

        return self._op_def

    @public
    @property
    def config(self) -> Any:
        """The config attached to the input that we're loading."""
        return self._config

    @public
    @property
    def metadata(self) -> Optional[ArbitraryMetadataMapping]:
        """A dict of metadata that is assigned to the InputDefinition that we're loading for.
        This property only contains metadata passed in explicitly with :py:class:`AssetIn`
        or :py:class:`In`. To access metadata of an upstream asset or operation definition,
        use the metadata in :py:attr:`.InputContext.upstream_output`.
        """
        return self._metadata

    @public
    @property
    def upstream_output(self) -> Optional["OutputContext"]:
        """Info about the output that produced the object we're loading."""
        return self._upstream_output

    @public
    @property
    def dagster_type(self) -> "DagsterType":
        """The type of this input.
        Dagster types do not propagate from an upstream output to downstream inputs,
        and this property only captures type information for the input that is either
        passed in explicitly with :py:class:`AssetIn` or :py:class:`In`, or can be
        infered from type hints. For an asset input, the Dagster type from the upstream
        asset definition is ignored.
        """
        if self._dagster_type is None:
            raise DagsterInvariantViolationError(
                "Attempting to access dagster_type, "
                "but it was not provided when constructing the InputContext"
            )

        return self._dagster_type

    @public
    @property
    def log(self) -> "DagsterLogManager":
        """The log manager to use for this input."""
        if self._log is None:
            raise DagsterInvariantViolationError(
                "Attempting to access log, "
                "but it was not provided when constructing the InputContext"
            )

        return self._log

    @public
    @property
    def resource_config(self) -> Optional[Mapping[str, Any]]:
        """The config associated with the resource that initializes the RootInputManager."""
        return self._resource_config

    @public
    @property
    def resources(self) -> Any:
        """The resources required by the resource that initializes the
        input manager. If using the :py:func:`@root_input_manager` decorator, these resources
        correspond to those requested with the `required_resource_keys` parameter.
        """
        if self._resources is None:
            raise DagsterInvariantViolationError(
                "Attempting to access resources, "
                "but it was not provided when constructing the InputContext"
            )

        if self._resources_cm and self._resources_contain_cm and not self._cm_scope_entered:
            raise DagsterInvariantViolationError(
                "At least one provided resource is a generator, but attempting to access "
                "resources outside of context manager scope. You can use the following syntax to "
                "open a context manager: `with build_input_context(...) as context:`"
            )
        return self._resources

    @public
    @property
    def has_asset_key(self) -> bool:
        return self._asset_key is not None

    @public
    @property
    def asset_key(self) -> AssetKey:
        if self._asset_key is None:
            raise DagsterInvariantViolationError(
                "Attempting to access asset_key, but no asset is associated with this input"
            )

        return self._asset_key

    @public
    @property
    def asset_partitions_def(self) -> "PartitionsDefinition":
        """The PartitionsDefinition on the upstream asset corresponding to this input."""
        if self._asset_partitions_def is None:
            if self.asset_key:
                raise DagsterInvariantViolationError(
                    f"Attempting to access partitions def for asset {self.asset_key}, but it is not"
                    " partitioned"
                )
            else:
                raise DagsterInvariantViolationError(
                    "Attempting to access partitions def for asset, but input does not correspond"
                    " to an asset"
                )

        return self._asset_partitions_def

    @property
    def step_context(self) -> "StepExecutionContext":
        if self._step_context is None:
            raise DagsterInvariantViolationError(
                "Attempting to access step_context, "
                "but it was not provided when constructing the InputContext"
            )

        return self._step_context

    @public
    @property
    def has_partition_key(self) -> bool:
        """Whether the current run is a partitioned run."""
        return self._partition_key is not None

    @public
    @property
    def partition_key(self) -> str:
        """The partition key for the current run.

        Raises an error if the current run is not a partitioned run.
        """
        if self._partition_key is None:
            check.failed(
                "Tried to access partition_key on a non-partitioned run.",
            )

        return self._partition_key

    @public
    @property
    def has_asset_partitions(self) -> bool:
        return self._asset_partitions_subset is not None

    @public
    @property
    def asset_partition_key(self) -> str:
        """The partition key for input asset.

        Raises an error if the input asset has no partitioning, or if the run covers a partition
        range for the input asset.
        """
        subset = self._asset_partitions_subset

        if subset is None:
            check.failed("The input does not correspond to a partitioned asset.")

        partition_keys = list(subset.get_partition_keys())
        if len(partition_keys) == 1:
            return partition_keys[0]
        else:
            check.failed(
                f"Tried to access partition key for asset '{self.asset_key}', "
                f"but the number of input partitions != 1: '{subset}'."
            )

    @public
    @property
    def asset_partition_key_range(self) -> PartitionKeyRange:
        """The partition key range for input asset.

        Raises an error if the input asset has no partitioning.
        """
        subset = self._asset_partitions_subset

        if subset is None:
            check.failed(
                "Tried to access asset_partition_key_range, but the asset is not partitioned.",
            )

        partition_key_ranges = subset.get_partition_key_ranges(
            dynamic_partitions_store=self.instance
        )
        if len(partition_key_ranges) != 1:
            check.failed(
                (
                    "Tried to access asset_partition_key_range, but there are "
                    f"({len(partition_key_ranges)}) key ranges associated with this input."
                ),
            )

        return partition_key_ranges[0]

    @public
    @property
    def asset_partition_keys(self) -> Sequence[str]:
        """The partition keys for input asset.

        Raises an error if the input asset has no partitioning.
        """
        if self._asset_partitions_subset is None:
            check.failed(
                "Tried to access asset_partition_keys, but the asset is not partitioned.",
            )

        return list(self._asset_partitions_subset.get_partition_keys())

    @public
    @property
    def asset_partitions_time_window(self) -> TimeWindow:
        """The time window for the partitions of the input asset.

        Raises an error if either of the following are true:
        - The input asset has no partitioning.
        - The input asset is not partitioned with a TimeWindowPartitionsDefinition.
        """
        subset = self._asset_partitions_subset

        if subset is None:
            check.failed(
                "Tried to access asset_partitions_time_window, but the asset is not partitioned.",
            )

        if not isinstance(subset, TimeWindowPartitionsSubset):
            check.failed(
                (
                    "Tried to access asset_partitions_time_window, but the asset is not partitioned"
                    " with time windows."
                ),
            )

        time_windows = subset.included_time_windows
        if len(time_windows) != 1:
            check.failed(
                (
                    "Tried to access asset_partitions_time_window, but there are "
                    f"({len(time_windows)}) time windows associated with this input."
                ),
            )

        return time_windows[0]

    @public
    def get_identifier(self) -> Sequence[str]:
        """Utility method to get a collection of identifiers that as a whole represent a unique
        step input.

        If not using memoization, the unique identifier collection consists of

        - ``run_id``: the id of the run which generates the input.
            Note: This method also handles the re-execution memoization logic. If the step that
            generates the input is skipped in the re-execution, the ``run_id`` will be the id
            of its parent run.
        - ``step_key``: the key for a compute step.
        - ``name``: the name of the output. (default: 'result').

        If using memoization, the ``version`` corresponding to the step output is used in place of
        the ``run_id``.

        Returns:
            List[str, ...]: A list of identifiers, i.e. (run_id or version), step_key, and output_name
        """
        if self.upstream_output is None:
            raise DagsterInvariantViolationError(
                "InputContext.upstream_output not defined. Cannot compute an identifier"
            )

        return self.upstream_output.get_identifier()

    @public
    def get_asset_identifier(self) -> Sequence[str]:
        if self.asset_key is not None:
            if self.has_asset_partitions:
                return [*self.asset_key.path, self.asset_partition_key]
            else:
                return self.asset_key.path
        else:
            check.failed("Can't get asset identifier for an input with no asset key")

    def consume_events(self) -> Iterator["DagsterEvent"]:
        """Pops and yields all user-generated events that have been recorded from this context.

        If consume_events has not yet been called, this will yield all logged events since the call to `handle_input`. If consume_events has been called, it will yield all events since the last time consume_events was called. Designed for internal use. Users should never need to invoke this method.
        """
        events = self._events
        self._events = []
        yield from events

    def add_input_metadata(
        self,
        metadata: Mapping[str, Any],
        description: Optional[str] = None,
    ) -> None:
        """Accepts a dictionary of metadata. Metadata entries will appear on the LOADED_INPUT event.
        If the input is an asset, metadata will be attached to an asset observation.

        The asset observation will be yielded from the run and appear in the event log.
        Only valid if the context has an asset key.
        """
        from dagster._core.definitions.metadata import normalize_metadata
        from dagster._core.events import DagsterEvent

        metadata = check.mapping_param(metadata, "metadata", key_type=str)
        self._metadata = {**self._metadata, **normalize_metadata(metadata)}
        if self.has_asset_key:
            check.opt_str_param(description, "description")

            observation = AssetObservation(
                asset_key=self.asset_key,
                description=description,
                partition=self.asset_partition_key if self.has_asset_partitions else None,
                metadata=metadata,
            )
            self._observations.append(observation)
            if self._step_context:
                self._events.append(DagsterEvent.asset_observation(self._step_context, observation))

    def get_observations(
        self,
    ) -> Sequence[AssetObservation]:
        """Retrieve the list of user-generated asset observations that were observed via the context.

        User-generated events that were yielded will not appear in this list.

        **Examples:**

        .. code-block:: python

            from dagster import IOManager, build_input_context, AssetObservation

            class MyIOManager(IOManager):
                def load_input(self, context, obj):
                    ...

            def test_load_input():
                mgr = MyIOManager()
                context = build_input_context()
                mgr.load_input(context)
                observations = context.get_observations()
                ...
        """
        return self._observations

    def consume_metadata(self) -> Mapping[str, MetadataValue]:
        result = self._metadata
        self._metadata = {}
        return result


def build_input_context(
    name: Optional[str] = None,
    config: Optional[Any] = None,
    metadata: Optional[ArbitraryMetadataMapping] = None,
    upstream_output: Optional["OutputContext"] = None,
    dagster_type: Optional["DagsterType"] = None,
    resource_config: Optional[Mapping[str, Any]] = None,
    resources: Optional[Mapping[str, Any]] = None,
    op_def: Optional["OpDefinition"] = None,
    step_context: Optional["StepExecutionContext"] = None,
    asset_key: Optional[CoercibleToAssetKey] = None,
    partition_key: Optional[str] = None,
    asset_partition_key_range: Optional[PartitionKeyRange] = None,
    asset_partitions_def: Optional["PartitionsDefinition"] = None,
    instance: Optional[DagsterInstance] = None,
) -> "InputContext":
    """Builds input context from provided parameters.

    ``build_input_context`` can be used as either a function, or a context manager. If resources
    that are also context managers are provided, then ``build_input_context`` must be used as a
    context manager.

    Args:
        name (Optional[str]): The name of the input that we're loading.
        config (Optional[Any]): The config attached to the input that we're loading.
        metadata (Optional[Dict[str, Any]]): A dict of metadata that is assigned to the
            InputDefinition that we're loading for.
        upstream_output (Optional[OutputContext]): Info about the output that produced the object
            we're loading.
        dagster_type (Optional[DagsterType]): The type of this input.
        resource_config (Optional[Dict[str, Any]]): The resource config to make available from the
            input context. This usually corresponds to the config provided to the resource that
            loads the input manager.
        resources (Optional[Dict[str, Any]]): The resources to make available from the context.
            For a given key, you can provide either an actual instance of an object, or a resource
            definition.
        asset_key (Optional[Union[AssetKey, Sequence[str], str]]): The asset key attached to the InputDefinition.
        op_def (Optional[OpDefinition]): The definition of the op that's loading the input.
        step_context (Optional[StepExecutionContext]): For internal use.
        partition_key (Optional[str]): String value representing partition key to execute with.
        asset_partition_key_range (Optional[str]): The range of asset partition keys to load.
        asset_partitions_def: Optional[PartitionsDefinition]: The PartitionsDefinition of the asset
            being loaded.

    Examples:
        .. code-block:: python

            build_input_context()

            with build_input_context(resources={"foo": context_manager_resource}) as context:
                do_something
    """
    from dagster._core.definitions import OpDefinition, PartitionsDefinition
    from dagster._core.execution.context.output import OutputContext
    from dagster._core.execution.context.system import StepExecutionContext
    from dagster._core.execution.context_creation_job import initialize_console_manager
    from dagster._core.types.dagster_type import DagsterType

    name = check.opt_str_param(name, "name")
    metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
    upstream_output = check.opt_inst_param(upstream_output, "upstream_output", OutputContext)
    dagster_type = check.opt_inst_param(dagster_type, "dagster_type", DagsterType)
    resource_config = check.opt_mapping_param(resource_config, "resource_config", key_type=str)
    resources = check.opt_mapping_param(resources, "resources", key_type=str)
    op_def = check.opt_inst_param(op_def, "op_def", OpDefinition)
    step_context = check.opt_inst_param(step_context, "step_context", StepExecutionContext)
    asset_key = AssetKey.from_coercible(asset_key) if asset_key else None
    partition_key = check.opt_str_param(partition_key, "partition_key")
    asset_partition_key_range = check.opt_inst_param(
        asset_partition_key_range, "asset_partition_key_range", PartitionKeyRange
    )
    asset_partitions_def = check.opt_inst_param(
        asset_partitions_def, "asset_partitions_def", PartitionsDefinition
    )
    if asset_partitions_def and asset_partition_key_range:
        asset_partitions_subset = asset_partitions_def.empty_subset().with_partition_key_range(
            asset_partition_key_range, dynamic_partitions_store=instance
        )
    elif asset_partition_key_range:
        asset_partitions_subset = KeyRangeNoPartitionsDefPartitionsSubset(asset_partition_key_range)
    else:
        asset_partitions_subset = None

    return InputContext(
        name=name,
        job_name=None,
        config=config,
        metadata=metadata,
        upstream_output=upstream_output,
        dagster_type=dagster_type,
        log_manager=initialize_console_manager(None),
        resource_config=resource_config,
        resources=resources,
        step_context=step_context,
        op_def=op_def,
        asset_key=asset_key,
        partition_key=partition_key,
        asset_partitions_subset=asset_partitions_subset,
        asset_partitions_def=asset_partitions_def,
        instance=instance,
    )


class KeyRangeNoPartitionsDefPartitionsSubset(PartitionsSubset):
    """For build_input_context when no PartitionsDefinition has been provided."""

    def __init__(self, key_range: PartitionKeyRange):
        self._key_range = key_range

    def get_partition_keys_not_in_subset(
        self,
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> Iterable[str]:
        raise NotImplementedError()

    def get_partition_keys(self, current_time: Optional[datetime] = None) -> Iterable[str]:
        if self._key_range.start == self._key_range.end:
            return self._key_range.start
        else:
            raise NotImplementedError()

    def get_partition_key_ranges(
        self,
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> Sequence[PartitionKeyRange]:
        return [self._key_range]

    def with_partition_keys(self, partition_keys: Iterable[str]) -> "PartitionsSubset":
        raise NotImplementedError()

    def with_partition_key_range(
        self,
        partition_key_range: PartitionKeyRange,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> "PartitionsSubset":
        raise NotImplementedError()

    def serialize(self) -> str:
        raise NotImplementedError()

    @property
    def partitions_def(self) -> "PartitionsDefinition":
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()

    def __contains__(self, value) -> bool:
        raise NotImplementedError()

    @classmethod
    def from_serialized(
        cls, partitions_def: "PartitionsDefinition", serialized: str
    ) -> "PartitionsSubset":
        raise NotImplementedError()

    @classmethod
    def can_deserialize(
        cls,
        partitions_def: "PartitionsDefinition",
        serialized: str,
        serialized_partitions_def_unique_id: Optional[str],
        serialized_partitions_def_class_name: Optional[str],
    ) -> bool:
        raise NotImplementedError()

    @classmethod
    def empty_subset(cls, partitions_def: "PartitionsDefinition") -> "PartitionsSubset":
        raise NotImplementedError()



import warnings
from inspect import Parameter
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

import dagster._check as check
from dagster._builtins import Nothing
from dagster._config import UserConfigSchema
from dagster._core.decorator_utils import get_function_params, get_valid_name_permutations
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import ArbitraryMetadataMapping, MetadataUserInput
from dagster._core.definitions.resource_annotation import (
    get_resource_args,
)
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.types.dagster_type import DagsterType
from dagster._utils.backcompat import (
    ExperimentalWarning,
    deprecation_warning,
    experimental_arg_warning,
)

from ..asset_in import AssetIn
from ..asset_out import AssetOut
from ..assets import AssetsDefinition
from ..decorators.graph_decorator import graph
from ..decorators.op_decorator import _Op
from ..events import AssetKey, CoercibleToAssetKey, CoercibleToAssetKeyPrefix
from ..input import In
from ..output import GraphOut, Out
from ..partition import PartitionsDefinition
from ..policy import RetryPolicy
from ..resource_definition import ResourceDefinition
from ..utils import DEFAULT_IO_MANAGER_KEY, NoValueSentinel


@overload
def asset(
    compute_fn: Callable,
) -> AssetsDefinition:
    ...


@overload
def asset(
    *,
    name: Optional[str] = ...,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = ...,
    non_argument_deps: Optional[Union[Set[AssetKey], Set[str]]] = ...,
    metadata: Optional[Mapping[str, Any]] = ...,
    description: Optional[str] = ...,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[Set[str]] = ...,
    resource_defs: Optional[Mapping[str, object]] = ...,
    io_manager_def: Optional[object] = ...,
    io_manager_key: Optional[str] = ...,
    compute_kind: Optional[str] = ...,
    dagster_type: Optional[DagsterType] = ...,
    partitions_def: Optional[PartitionsDefinition] = ...,
    op_tags: Optional[Mapping[str, Any]] = ...,
    group_name: Optional[str] = ...,
    output_required: bool = ...,
    freshness_policy: Optional[FreshnessPolicy] = ...,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = ...,
    retry_policy: Optional[RetryPolicy] = ...,
    code_version: Optional[str] = ...,
    key: Optional[CoercibleToAssetKey] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    ...


def asset(
    compute_fn: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    non_argument_deps: Optional[Union[Set[AssetKey], Set[str]]] = None,
    metadata: Optional[ArbitraryMetadataMapping] = None,
    description: Optional[str] = None,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[Set[str]] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    io_manager_def: Optional[object] = None,
    io_manager_key: Optional[str] = None,
    compute_kind: Optional[str] = None,
    dagster_type: Optional[DagsterType] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    group_name: Optional[str] = None,
    output_required: bool = True,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    retry_policy: Optional[RetryPolicy] = None,
    code_version: Optional[str] = None,
    key: Optional[CoercibleToAssetKey] = None,
) -> Union[AssetsDefinition, Callable[[Callable[..., Any]], AssetsDefinition]]:
    """Create a definition for how to compute an asset.

    A software-defined asset is the combination of:
      1. An asset key, e.g. the name of a table.
      2. A function, which can be run to compute the contents of the asset.
      3. A set of upstream assets that are provided as inputs to the function when computing the asset.

    Unlike an op, whose dependencies are determined by the graph it lives inside, an asset knows
    about the upstream assets it depends on. The upstream assets are inferred from the arguments
    to the decorated function. The name of the argument designates the name of the upstream asset.

    An asset has an op inside it to represent the function that computes it. The name of the op
    will be the segments of the asset key, separated by double-underscores.

    Args:
        name (Optional[str]): The name of the asset.  If not provided, defaults to the name of the
            decorated function. The asset's name must be a valid name in dagster (ie only contains
            letters, numbers, and _) and may not contain python reserved keywords.
        key_prefix (Optional[Union[str, Sequence[str]]]): If provided, the asset's key is the
            concatenation of the key_prefix and the asset's name, which defaults to the name of
            the decorated function. Each item in key_prefix must be a valid name in dagster (ie only
            contains letters, numbers, and _) and may not contain python reserved keywords.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        non_argument_deps (Optional[Union[Set[AssetKey], Set[str]]]): Set of asset keys that are
            upstream dependencies, but do not pass an input to the asset.
        config_schema (Optional[ConfigSchema): The configuration schema for the asset's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        metadata (Optional[Dict[str, Any]]): A dict of metadata entries for the asset.
        required_resource_keys (Optional[Set[str]]): Set of resource handles required by the op.
        io_manager_key (Optional[str]): The resource key of the IOManager used
            for storing the output of the op as an asset, and for loading it in downstream ops
            (default: "io_manager"). Only one of io_manager_key and io_manager_def can be provided.
        io_manager_def (Optional[object]): (Experimental) The IOManager used for
            storing the output of the op as an asset,  and for loading it in
            downstream ops. Only one of io_manager_def and io_manager_key can be provided.
        compute_kind (Optional[str]): A string to represent the kind of computation that produces
            the asset, e.g. "dbt" or "spark". It will be displayed in Dagit as a badge on the asset.
        dagster_type (Optional[DagsterType]): Allows specifying type validation functions that
            will be executed on the output of the decorated function after it runs.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that computes the asset.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If not provided,
            the name "default" is used.
        resource_defs (Optional[Mapping[str, object]]):
            (Experimental) A mapping of resource keys to resources. These resources
            will be initialized during execution, and can be accessed from the
            context within the body of the function.
        output_required (bool): Whether the decorated function will always materialize an asset.
            Defaults to True. If False, the function can return None, which will not be materialized to
            storage and will halt execution of downstream assets.
        freshness_policy (FreshnessPolicy): A constraint telling Dagster how often this asset is intended to be updated
            with respect to its root data.
        auto_materialize_policy (AutoMaterializePolicy): (Experimental) Configure Dagster to automatically materialize
            this asset according to its FreshnessPolicy and when upstream dependencies change.
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that computes the asset.
        code_version (Optional[str]): (Experimental) Version of the code that generates this asset. In
            general, versions should be set only for code that deterministically produces the same
            output when given the same inputs.

    Examples:
        .. code-block:: python

            @asset
            def my_asset(my_upstream_asset: int) -> int:
                return my_upstream_asset + 1
    """

    def create_asset():
        return _Asset(
            name=cast(Optional[str], name),  # (mypy bug that it can't infer name is Optional[str])
            key_prefix=key_prefix,
            ins=ins,
            non_argument_deps=_make_asset_keys(non_argument_deps),
            metadata=metadata,
            description=description,
            config_schema=config_schema,
            required_resource_keys=required_resource_keys,
            resource_defs=resource_defs,
            io_manager_key=io_manager_key,
            io_manager_def=io_manager_def,
            compute_kind=check.opt_str_param(compute_kind, "compute_kind"),
            dagster_type=dagster_type,
            partitions_def=partitions_def,
            op_tags=op_tags,
            group_name=group_name,
            output_required=output_required,
            freshness_policy=freshness_policy,
            auto_materialize_policy=auto_materialize_policy,
            retry_policy=retry_policy,
            code_version=code_version,
            key=key,
        )

    if compute_fn is not None:
        return create_asset()(compute_fn)

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        check.invariant(
            not (io_manager_key and io_manager_def),
            (
                "Both io_manager_key and io_manager_def were provided to `@asset` decorator. Please"
                " provide one or the other. "
            ),
        )
        if resource_defs is not None:
            experimental_arg_warning("resource_defs", "asset")

        if io_manager_def is not None:
            experimental_arg_warning("io_manager_def", "asset")

        if auto_materialize_policy is not None:
            experimental_arg_warning("auto_materialize_policy", "asset")

        return create_asset()(fn)

    return inner


class _Asset:
    def __init__(
        self,
        name: Optional[str] = None,
        key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
        ins: Optional[Mapping[str, AssetIn]] = None,
        non_argument_deps: Optional[Set[AssetKey]] = None,
        metadata: Optional[ArbitraryMetadataMapping] = None,
        description: Optional[str] = None,
        config_schema: Optional[UserConfigSchema] = None,
        required_resource_keys: Optional[Set[str]] = None,
        resource_defs: Optional[Mapping[str, object]] = None,
        io_manager_key: Optional[str] = None,
        io_manager_def: Optional[object] = None,
        compute_kind: Optional[str] = None,
        dagster_type: Optional[DagsterType] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
        op_tags: Optional[Mapping[str, Any]] = None,
        group_name: Optional[str] = None,
        output_required: bool = True,
        freshness_policy: Optional[FreshnessPolicy] = None,
        auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
        retry_policy: Optional[RetryPolicy] = None,
        code_version: Optional[str] = None,
        key: Optional[CoercibleToAssetKey] = None,
    ):
        self.name = name

        if isinstance(key_prefix, str):
            key_prefix = [key_prefix]
        self.key_prefix = key_prefix
        self.ins = ins or {}
        self.non_argument_deps = non_argument_deps
        self.metadata = metadata
        self.description = description
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys"
        )
        self.io_manager_key = io_manager_key
        self.io_manager_def = io_manager_def
        self.config_schema = config_schema
        self.compute_kind = compute_kind
        self.dagster_type = dagster_type
        self.partitions_def = partitions_def
        self.op_tags = op_tags
        self.resource_defs = dict(check.opt_mapping_param(resource_defs, "resource_defs"))
        self.group_name = group_name
        self.output_required = output_required
        self.freshness_policy = freshness_policy
        self.retry_policy = retry_policy
        self.auto_materialize_policy = auto_materialize_policy
        self.code_version = code_version

        if (name or key_prefix) and key:
            raise DagsterInvalidDefinitionError(
                "Cannot specify a name or key prefix for an asset when the key argument is"
                " provided."
            )

        self.key = AssetKey.from_coercible(key) if key is not None else None

    def __call__(self, fn: Callable) -> AssetsDefinition:
        from dagster._config.pythonic_config import (
            validate_resource_annotated_function,
        )
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        validate_resource_annotated_function(fn)
        asset_name = self.name or fn.__name__

        asset_ins = build_asset_ins(fn, self.ins or {}, self.non_argument_deps)

        out_asset_key = (
            AssetKey(list(filter(None, [*(self.key_prefix or []), asset_name])))
            if not self.key
            else self.key
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ExperimentalWarning)

            arg_resource_keys = {arg.name for arg in get_resource_args(fn)}

            bare_required_resource_keys = set(self.required_resource_keys)

            resource_defs_dict = self.resource_defs
            resource_defs_keys = set(resource_defs_dict.keys())
            decorator_resource_keys = bare_required_resource_keys | resource_defs_keys

            io_manager_key = self.io_manager_key
            if self.io_manager_def:
                if not io_manager_key:
                    io_manager_key = out_asset_key.to_python_identifier("io_manager")

                if (
                    io_manager_key in self.resource_defs
                    and self.resource_defs[io_manager_key] != self.io_manager_def
                ):
                    raise DagsterInvalidDefinitionError(
                        f"Provided conflicting definitions for io manager key '{io_manager_key}'."
                        " Please provide only one definition per key."
                    )

                resource_defs_dict[io_manager_key] = self.io_manager_def

            wrapped_resource_defs = wrap_resources_for_execution(resource_defs_dict)

            check.param_invariant(
                len(bare_required_resource_keys) == 0 or len(arg_resource_keys) == 0,
                (
                    "Cannot specify resource requirements in both @asset decorator and as arguments"
                    " to the decorated function"
                ),
            )

            io_manager_key = cast(str, io_manager_key) if io_manager_key else DEFAULT_IO_MANAGER_KEY

            out = Out(
                metadata=self.metadata or {},
                io_manager_key=io_manager_key,
                dagster_type=self.dagster_type if self.dagster_type else NoValueSentinel,
                description=self.description,
                is_required=self.output_required,
                code_version=self.code_version,
            )

            op_required_resource_keys = decorator_resource_keys - arg_resource_keys

            op = _Op(
                name=out_asset_key.to_python_identifier(),
                description=self.description,
                ins=dict(asset_ins.values()),
                out=out,
                # Any resource requirements specified as arguments will be identified as
                # part of the Op definition instantiation
                required_resource_keys=op_required_resource_keys,
                tags={
                    **({"kind": self.compute_kind} if self.compute_kind else {}),
                    **(self.op_tags or {}),
                },
                config_schema=self.config_schema,
                retry_policy=self.retry_policy,
                code_version=self.code_version,
            )(fn)

        keys_by_input_name = {
            input_name: asset_key for asset_key, (input_name, _) in asset_ins.items()
        }
        partition_mappings = {
            keys_by_input_name[input_name]: asset_in.partition_mapping
            for input_name, asset_in in self.ins.items()
            if asset_in.partition_mapping is not None
        }

        return AssetsDefinition.dagster_internal_init(
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={"result": out_asset_key},
            node_def=op,
            partitions_def=self.partitions_def,
            partition_mappings=partition_mappings if partition_mappings else None,
            resource_defs=wrapped_resource_defs,
            group_names_by_key={out_asset_key: self.group_name} if self.group_name else None,
            freshness_policies_by_key={out_asset_key: self.freshness_policy}
            if self.freshness_policy
            else None,
            auto_materialize_policies_by_key={out_asset_key: self.auto_materialize_policy}
            if self.auto_materialize_policy
            else None,
            asset_deps=None,  # no asset deps in single-asset decorator
            selected_asset_keys=None,  # no subselection in decorator
            can_subset=False,
            metadata_by_key=None,  # not supported for now
            descriptions_by_key=None,  # not supported for now
        )


def multi_asset(
    *,
    outs: Mapping[str, AssetOut],
    name: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    non_argument_deps: Optional[Union[Set[AssetKey], Set[str]]] = None,
    description: Optional[str] = None,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[Set[str]] = None,
    compute_kind: Optional[str] = None,
    internal_asset_deps: Optional[Mapping[str, Set[AssetKey]]] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    can_subset: bool = False,
    resource_defs: Optional[Mapping[str, object]] = None,
    group_name: Optional[str] = None,
    retry_policy: Optional[RetryPolicy] = None,
    code_version: Optional[str] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    """Create a combined definition of multiple assets that are computed using the same op and same
    upstream assets.

    Each argument to the decorated function references an upstream asset that this asset depends on.
    The name of the argument designates the name of the upstream asset.

    You can set I/O managers keys, auto-materialize policies, freshness policies, group names, etc.
    on an individual asset within the multi-asset by attaching them to the :py:class:`AssetOut`
    corresponding to that asset in the `outs` parameter.

    Args:
        name (Optional[str]): The name of the op.
        outs: (Optional[Dict[str, AssetOut]]): The AssetOuts representing the produced assets.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        non_argument_deps (Optional[Union[Set[AssetKey], Set[str]]]): Set of asset keys that are upstream
            dependencies, but do not pass an input to the multi_asset.
        config_schema (Optional[ConfigSchema): The configuration schema for the asset's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        required_resource_keys (Optional[Set[str]]): Set of resource handles required by the underlying op.
        compute_kind (Optional[str]): A string to represent the kind of computation that produces
            the asset, e.g. "dbt" or "spark". It will be displayed in Dagit as a badge on the asset.
        internal_asset_deps (Optional[Mapping[str, Set[AssetKey]]]): By default, it is assumed
            that all assets produced by a multi_asset depend on all assets that are consumed by that
            multi asset. If this default is not correct, you pass in a map of output names to a
            corrected set of AssetKeys that they depend on. Any AssetKeys in this list must be either
            used as input to the asset or produced within the op.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the assets.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that computes the asset.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        can_subset (bool): If this asset's computation can emit a subset of the asset
            keys based on the context.selected_assets argument. Defaults to False.
        resource_defs (Optional[Mapping[str, object]]):
            (Experimental) A mapping of resource keys to resources. These resources
            will be initialized during execution, and can be accessed from the
            context within the body of the function.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. This
            group name will be applied to all assets produced by this multi_asset.
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that computes the asset.
        code_version (Optional[str]): (Experimental) Version of the code encapsulated by the multi-asset. If set,
            this is used as a default code version for all defined assets.

    Examples:
        .. code-block:: python

            # Use IO managers to handle I/O:
            @multi_asset(
                outs={
                    "my_string_asset": AssetOut(),
                    "my_int_asset": AssetOut(),
                }
            )
            def my_function(upstream_asset: int):
                result = upstream_asset + 1
                return str(result), result

            # Handle I/O on your own:
            @multi_asset(
                outs={
                    "asset1": AssetOut(),
                    "asset2": AssetOut(),
                },
                non_argument_deps={"asset0"},
            )
            def my_function():
                asset0_value = load(path="asset0")
                asset1_result, asset2_result = do_some_transformation(asset0_value)
                write(asset1_result, path="asset1")
                write(asset2_result, path="asset2")
    """
    from dagster._core.execution.build_resources import wrap_resources_for_execution

    if resource_defs is not None:
        experimental_arg_warning("resource_defs", "multi_asset")

    asset_deps = check.opt_mapping_param(
        internal_asset_deps, "internal_asset_deps", key_type=str, value_type=set
    )
    required_resource_keys = check.opt_set_param(
        required_resource_keys, "required_resource_keys", of_type=str
    )
    resource_defs = wrap_resources_for_execution(
        check.opt_mapping_param(resource_defs, "resource_defs", key_type=str)
    )

    _config_schema = check.opt_mapping_param(
        config_schema,
        "config_schema",
        additional_message="Only dicts are supported for asset config_schema.",
    )

    bare_required_resource_keys = set(required_resource_keys)
    resource_defs_keys = set(resource_defs.keys())
    required_resource_keys = bare_required_resource_keys | resource_defs_keys

    for out in outs.values():
        if isinstance(out, Out) and not isinstance(out, AssetOut):
            deprecation_warning(
                "Passing Out objects as values for the out argument of @multi_asset",
                "1.0.0",
                additional_warn_txt="Use AssetOut instead.",
            )

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        op_name = name or fn.__name__
        asset_ins = build_asset_ins(
            fn, ins or {}, non_argument_deps=_make_asset_keys(non_argument_deps)
        )
        asset_outs = build_asset_outs(outs)

        arg_resource_keys = {arg.name for arg in get_resource_args(fn)}
        check.param_invariant(
            len(bare_required_resource_keys or []) == 0 or len(arg_resource_keys) == 0,
            (
                "Cannot specify resource requirements in both @multi_asset decorator and as"
                " arguments to the decorated function"
            ),
        )

        # validate that the asset_deps make sense
        valid_asset_deps = set(asset_ins.keys()) | set(asset_outs.keys())
        for out_name, asset_keys in asset_deps.items():
            check.invariant(
                out_name in outs,
                (
                    f"Invalid out key '{out_name}' supplied to `internal_asset_deps` argument for"
                    f" multi-asset {op_name}. Must be one of the outs for this multi-asset"
                    f" {list(outs.keys())[:20]}."
                ),
            )
            invalid_asset_deps = asset_keys.difference(valid_asset_deps)
            check.invariant(
                not invalid_asset_deps,
                (
                    f"Invalid asset dependencies: {invalid_asset_deps} specified in"
                    f" `internal_asset_deps` argument for multi-asset '{op_name}' on key"
                    f" '{out_name}'. Each specified asset key must be associated with an input to"
                    " the asset or produced by this asset. Valid keys:"
                    f" {list(valid_asset_deps)[:20]}"
                ),
            )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ExperimentalWarning)

            op_required_resource_keys = required_resource_keys - arg_resource_keys

            op = _Op(
                name=op_name,
                description=description,
                ins=dict(asset_ins.values()),
                out=dict(asset_outs.values()),
                required_resource_keys=op_required_resource_keys,
                tags={
                    **({"kind": compute_kind} if compute_kind else {}),
                    **(op_tags or {}),
                },
                config_schema=_config_schema,
                retry_policy=retry_policy,
                code_version=code_version,
            )(fn)

        keys_by_input_name = {
            input_name: asset_key for asset_key, (input_name, _) in asset_ins.items()
        }
        keys_by_output_name = {
            output_name: asset_key for asset_key, (output_name, _) in asset_outs.items()
        }

        # source group names from the AssetOuts (if any)
        group_names_by_key = {
            keys_by_output_name[output_name]: out.group_name
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.group_name is not None
        }
        if group_name:
            check.invariant(
                not group_names_by_key,
                (
                    "Cannot set group_name parameter on multi_asset if one or more of the AssetOuts"
                    " supplied to this multi_asset have a group_name defined."
                ),
            )
            group_names_by_key = {
                asset_key: group_name for asset_key in keys_by_output_name.values()
            }

        # source freshness policies from the AssetOuts (if any)
        freshness_policies_by_key = {
            keys_by_output_name[output_name]: out.freshness_policy
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.freshness_policy is not None
        }
        auto_materialize_policies_by_key = {
            keys_by_output_name[output_name]: out.auto_materialize_policy
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.auto_materialize_policy is not None
        }
        partition_mappings = {
            keys_by_input_name[input_name]: asset_in.partition_mapping
            for input_name, asset_in in (ins or {}).items()
            if asset_in.partition_mapping is not None
        }

        return AssetsDefinition.dagster_internal_init(
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name=keys_by_output_name,
            node_def=op,
            asset_deps={keys_by_output_name[name]: asset_deps[name] for name in asset_deps},
            partitions_def=partitions_def,
            partition_mappings=partition_mappings if partition_mappings else None,
            can_subset=can_subset,
            resource_defs=resource_defs,
            group_names_by_key=group_names_by_key,
            freshness_policies_by_key=freshness_policies_by_key,
            auto_materialize_policies_by_key=auto_materialize_policies_by_key,
            selected_asset_keys=None,  # no subselection in decorator
            descriptions_by_key=None,  # not supported for now
            metadata_by_key=None,  # not supported for now
        )

    return inner


def build_asset_ins(
    fn: Callable,
    asset_ins: Mapping[str, AssetIn],
    non_argument_deps: Optional[AbstractSet[AssetKey]],
) -> Mapping[AssetKey, Tuple[str, In]]:
    """Creates a mapping from AssetKey to (name of input, In object)."""
    non_argument_deps = check.opt_set_param(non_argument_deps, "non_argument_deps", AssetKey)

    params = get_function_params(fn)
    is_context_provided = len(params) > 0 and params[0].name in get_valid_name_permutations(
        "context"
    )
    input_params = params[1:] if is_context_provided else params

    # Filter config, resource args
    resource_arg_names = {arg.name for arg in get_resource_args(fn)}

    new_input_args = []
    for input_arg in input_params:
        if input_arg.name != "config" and input_arg.name not in resource_arg_names:
            new_input_args.append(input_arg)
    input_params = new_input_args

    non_var_input_param_names = [
        param.name for param in new_input_args if param.kind == Parameter.POSITIONAL_OR_KEYWORD
    ]
    has_kwargs = any(param.kind == Parameter.VAR_KEYWORD for param in new_input_args)

    all_input_names = set(non_var_input_param_names) | asset_ins.keys()

    if not has_kwargs:
        for in_key, asset_in in asset_ins.items():
            if in_key not in non_var_input_param_names and (
                not isinstance(asset_in.dagster_type, DagsterType)
                or not asset_in.dagster_type.is_nothing
            ):
                raise DagsterInvalidDefinitionError(
                    f"Key '{in_key}' in provided ins dict does not correspond to any of the names "
                    "of the arguments to the decorated function"
                )

    ins_by_asset_key: Dict[AssetKey, Tuple[str, In]] = {}
    for input_name in all_input_names:
        asset_key = None

        if input_name in asset_ins:
            asset_key = asset_ins[input_name].key
            metadata = asset_ins[input_name].metadata or {}
            key_prefix = asset_ins[input_name].key_prefix
            input_manager_key = asset_ins[input_name].input_manager_key
            dagster_type = asset_ins[input_name].dagster_type
        else:
            metadata = {}
            key_prefix = None
            input_manager_key = None
            dagster_type = NoValueSentinel

        asset_key = asset_key or AssetKey(list(filter(None, [*(key_prefix or []), input_name])))

        ins_by_asset_key[asset_key] = (
            input_name.replace("-", "_"),
            In(metadata=metadata, input_manager_key=input_manager_key, dagster_type=dagster_type),
        )

    for asset_key in non_argument_deps:
        stringified_asset_key = "_".join(asset_key.path).replace("-", "_")
        # mypy doesn't realize that Nothing is a valid type here
        ins_by_asset_key[asset_key] = (stringified_asset_key, In(cast(type, Nothing)))

    return ins_by_asset_key


@overload
def graph_asset(compose_fn: Callable) -> AssetsDefinition:
    ...


@overload
def graph_asset(
    *,
    name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    description: Optional[str] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    group_name: Optional[str] = None,
    metadata: Optional[MetadataUserInput] = ...,
    freshness_policy: Optional[FreshnessPolicy] = ...,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = ...,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    ...


def graph_asset(
    compose_fn: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    description: Optional[str] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    group_name: Optional[str] = None,
    metadata: Optional[MetadataUserInput] = None,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
) -> Union[AssetsDefinition, Callable[[Callable[..., Any]], AssetsDefinition]]:
    """Creates a software-defined asset that's computed using a graph of ops.

    This decorator is meant to decorate a function that composes a set of ops or graphs to define
    the dependencies between them.

    Args:
        name (Optional[str]): The name of the asset.  If not provided, defaults to the name of the
            decorated function. The asset's name must be a valid name in Dagster (ie only contains
            letters, numbers, and underscores) and may not contain Python reserved keywords.
        key_prefix (Optional[Union[str, Sequence[str]]]): If provided, the asset's key is the
            concatenation of the key_prefix and the asset's name, which defaults to the name of
            the decorated function. Each item in key_prefix must be a valid name in Dagster (ie only
            contains letters, numbers, and underscores) and may not contain Python reserved keywords.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If
            not provided, the name "default" is used.
        metadata (Optional[MetadataUserInput]): Dictionary of metadata to be associated with
            the asset.
        freshness_policy (Optional[FreshnessPolicy]): A constraint telling Dagster how often this asset is
            intended to be updated with respect to its root data.
        auto_materialize_policy (Optional[AutoMaterializePolicy]): The AutoMaterializePolicy to use
            for this asset.

    Examples:
        .. code-block:: python

            @op
            def fetch_files_from_slack(context) -> pd.DataFrame:
                ...

            @op
            def store_files_in_table(files) -> None:
                files.to_sql(name="slack_files", con=create_db_connection())

            @graph_asset
            def slack_files_table():
                return store_files(fetch_files_from_slack())
    """
    if compose_fn is not None:
        return _GraphBackedAsset()(compose_fn)

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        return _GraphBackedAsset(
            name=cast(Optional[str], name),  # (mypy bug that it can't infer name is Optional[str])
            key_prefix=key_prefix,
            ins=ins,
            description=description,
            partitions_def=partitions_def,
            group_name=group_name,
            metadata=metadata,
            freshness_policy=freshness_policy,
            auto_materialize_policy=auto_materialize_policy,
            resource_defs=resource_defs,
        )(fn)

    return inner


class _GraphBackedAsset:
    def __init__(
        self,
        name: Optional[str] = None,
        key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
        ins: Optional[Mapping[str, AssetIn]] = None,
        description: Optional[str] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
        group_name: Optional[str] = None,
        metadata: Optional[MetadataUserInput] = None,
        freshness_policy: Optional[FreshnessPolicy] = None,
        auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    ):
        self.name = name

        if isinstance(key_prefix, str):
            key_prefix = [key_prefix]
        self.key_prefix = key_prefix
        self.ins = ins or {}
        self.description = description
        self.partitions_def = partitions_def
        self.group_name = group_name
        self.metadata = metadata
        self.freshness_policy = freshness_policy
        self.auto_materialize_policy = auto_materialize_policy
        self.resource_defs = resource_defs

    def __call__(self, fn: Callable) -> AssetsDefinition:
        asset_name = self.name or fn.__name__
        asset_ins = build_asset_ins(fn, self.ins or {}, set())
        out_asset_key = AssetKey(list(filter(None, [*(self.key_prefix or []), asset_name])))

        keys_by_input_name = {
            input_name: asset_key for asset_key, (input_name, _) in asset_ins.items()
        }
        partition_mappings = {
            input_name: asset_in.partition_mapping
            for input_name, asset_in in self.ins.items()
            if asset_in.partition_mapping
        }

        op_graph = graph(name=out_asset_key.to_python_identifier(), description=self.description)(
            fn
        )
        return AssetsDefinition.from_graph(
            op_graph,
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={"result": out_asset_key},
            partitions_def=self.partitions_def,
            partition_mappings=partition_mappings if partition_mappings else None,
            group_name=self.group_name,
            metadata_by_output_name={"result": self.metadata} if self.metadata else None,
            freshness_policies_by_output_name={"result": self.freshness_policy}
            if self.freshness_policy
            else None,
            auto_materialize_policies_by_output_name={"result": self.auto_materialize_policy}
            if self.auto_materialize_policy
            else None,
            descriptions_by_output_name={"result": self.description} if self.description else None,
            resource_defs=self.resource_defs,
        )


def graph_multi_asset(
    *,
    outs: Mapping[str, AssetOut],
    name: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    group_name: Optional[str] = None,
    can_subset: bool = False,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    """Create a combined definition of multiple assets that are computed using the same graph of
    ops, and the same upstream assets.

    Each argument to the decorated function references an upstream asset that this asset depends on.
    The name of the argument designates the name of the upstream asset.

    Args:
        name (Optional[str]): The name of the graph.
        outs: (Optional[Dict[str, AssetOut]]): The AssetOuts representing the produced assets.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the assets.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. This
            group name will be applied to all assets produced by this multi_asset.
        can_subset (bool): Whether this asset's computation can emit a subset of the asset
            keys based on the context.selected_assets argument. Defaults to False.
    """

    def inner(fn: Callable) -> AssetsDefinition:
        partition_mappings = {
            input_name: asset_in.partition_mapping
            for input_name, asset_in in (ins or {}).items()
            if asset_in.partition_mapping
        }

        asset_ins = build_asset_ins(fn, ins or {}, set())
        keys_by_input_name = {
            input_name: asset_key for asset_key, (input_name, _) in asset_ins.items()
        }
        asset_outs = build_asset_outs(outs)
        op_graph = graph(
            name=name or fn.__name__,
            out={out_name: GraphOut() for out_name, _ in asset_outs.values()},
        )(fn)

        # source metadata from the AssetOuts (if any)
        metadata_by_output_name = {
            output_name: out.metadata
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.metadata is not None
        }

        # source freshness policies from the AssetOuts (if any)
        freshness_policies_by_output_name = {
            output_name: out.freshness_policy
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.freshness_policy is not None
        }

        # source auto materialize policies from the AssetOuts (if any)
        auto_materialize_policies_by_output_name = {
            output_name: out.auto_materialize_policy
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.auto_materialize_policy is not None
        }

        # source descriptions from the AssetOuts (if any)
        descriptions_by_output_name = {
            output_name: out.description
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.description is not None
        }

        return AssetsDefinition.from_graph(
            op_graph,
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={
                output_name: asset_key for asset_key, (output_name, _) in asset_outs.items()
            },
            partitions_def=partitions_def,
            partition_mappings=partition_mappings if partition_mappings else None,
            group_name=group_name,
            can_subset=can_subset,
            metadata_by_output_name=metadata_by_output_name,
            freshness_policies_by_output_name=freshness_policies_by_output_name,
            auto_materialize_policies_by_output_name=auto_materialize_policies_by_output_name,
            descriptions_by_output_name=descriptions_by_output_name,
            resource_defs=resource_defs,
        )

    return inner


def build_asset_outs(asset_outs: Mapping[str, AssetOut]) -> Mapping[AssetKey, Tuple[str, Out]]:
    """Creates a mapping from AssetKey to (name of output, Out object)."""
    outs_by_asset_key: Dict[AssetKey, Tuple[str, Out]] = {}
    for output_name, asset_out in asset_outs.items():
        out = asset_out.to_out()
        asset_key = asset_out.key or AssetKey(
            list(filter(None, [*(asset_out.key_prefix or []), output_name]))
        )

        outs_by_asset_key[asset_key] = (output_name.replace("-", "_"), out)

    return outs_by_asset_key


def _make_asset_keys(deps: Optional[Union[Set[AssetKey], Set[str]]]) -> Optional[Set[AssetKey]]:
    """Convert all str items to AssetKey in the set."""
    if deps is None:
        return deps

    deps_asset_keys = {
        AssetKey.from_user_string(dep) if isinstance(dep, str) else dep for dep in deps
    }
    return deps_asset_keys



from collections import defaultdict
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)

import pendulum

import dagster._check as check
from dagster._core.definitions.asset_graph import AssetGraph
from dagster._core.definitions.data_version import (
    DATA_VERSION_TAG,
    DataVersion,
    extract_data_version_from_entry,
)
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.partition import DynamicPartitionsDefinition
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._core.events import DagsterEventType
from dagster._core.instance import DagsterInstance, DynamicPartitionsStore
from dagster._core.storage.dagster_run import (
    DagsterRun,
    RunRecord,
)
from dagster._core.storage.tags import PARTITION_NAME_TAG
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.storage.event_log import EventLogRecord
    from dagster._core.storage.event_log.base import AssetRecord


class CachingInstanceQueryer(DynamicPartitionsStore):
    """Provides utility functions for querying for asset-materialization related data from the
    instance which will attempt to limit redundant expensive calls. Intended for use within the
    scope of a single "request" (e.g. GQL request, sensor tick).

    Args:
        instance (DagsterInstance): The instance to query.
    """

    def __init__(
        self,
        instance: DagsterInstance,
        asset_graph: AssetGraph,
        evaluation_time: Optional[datetime] = None,
    ):
        self._instance = instance
        self._asset_graph = asset_graph

        self._asset_record_cache: Dict[AssetKey, Optional[AssetRecord]] = {}
        self._asset_partition_count_cache: Dict[
            Optional[int], Dict[AssetKey, Mapping[str, int]]
        ] = defaultdict(dict)

        self._dynamic_partitions_cache: Dict[str, Sequence[str]] = {}

        self._evaluation_time = evaluation_time if evaluation_time else pendulum.now("UTC")

    @property
    def instance(self) -> DagsterInstance:
        return self._instance

    @property
    def asset_graph(self) -> AssetGraph:
        return self._asset_graph

    @property
    def evaluation_time(self) -> datetime:
        return self._evaluation_time

    ####################
    # QUERY BATCHING
    ####################

    def prefetch_asset_partition_counts(
        self, asset_keys: Sequence[AssetKey], after_cursor: Optional[int]
    ):
        """For performance, batches together queries for selected assets."""
        if after_cursor is not None:
            self._asset_partition_count_cache[after_cursor] = dict(
                self.instance.get_materialization_count_by_partition(
                    asset_keys=asset_keys,
                    after_cursor=after_cursor,
                )
            )

    def prefetch_asset_records(self, asset_keys: Sequence[AssetKey]):
        """For performance, batches together queries for selected assets."""
        # get all asset records for the selected assets
        asset_records = self.instance.get_asset_records(asset_keys)
        for asset_record in asset_records:
            self._asset_record_cache[asset_record.asset_entry.asset_key] = asset_record
        for key in asset_keys:
            if key not in self._asset_record_cache:
                self._asset_record_cache[key] = None

    ####################
    # ASSET RECORDS / STORAGE IDS
    ####################

    def get_asset_record(self, asset_key: AssetKey) -> Optional["AssetRecord"]:
        if asset_key not in self._asset_record_cache:
            self._asset_record_cache[asset_key] = next(
                iter(self.instance.get_asset_records([asset_key])), None
            )
        return self._asset_record_cache[asset_key]

    def _event_type_for_key(self, asset_key: AssetKey) -> DagsterEventType:
        if self.asset_graph.is_source(asset_key):
            return DagsterEventType.ASSET_OBSERVATION
        else:
            return DagsterEventType.ASSET_MATERIALIZATION

    @cached_method
    def _get_latest_materialization_or_observation_record(
        self, *, asset_partition: AssetKeyPartitionKey, before_cursor: Optional[int] = None
    ) -> Optional["EventLogRecord"]:
        """Returns the latest event log record for the given asset partition of an asset. For
        observable source assets, this will be an AssetObservation, otherwise it will be an
        AssetMaterialization.
        """
        from dagster._core.event_api import EventRecordsFilter

        # in the simple case, just use the asset record
        if (
            before_cursor is None
            and asset_partition.partition_key is None
            and not self.asset_graph.is_observable(asset_partition.asset_key)
        ):
            asset_record = self.get_asset_record(asset_partition.asset_key)
            if asset_record is None:
                return None
            return asset_record.asset_entry.last_materialization_record

        records = self.instance.get_event_records(
            EventRecordsFilter(
                event_type=self._event_type_for_key(asset_partition.asset_key),
                asset_key=asset_partition.asset_key,
                asset_partitions=[asset_partition.partition_key]
                if asset_partition.partition_key
                else None,
                before_cursor=before_cursor,
            ),
            ascending=False,
            limit=1,
        )
        return next(iter(records), None)

    @cached_method
    def get_latest_storage_id_for_event_type(
        self, *, event_type: DagsterEventType
    ) -> Optional[int]:
        """Returns the latest storage id across all events of the given event_type.

        Args:
            event_type (DagsterEventType): The event type to query for.
        """
        from dagster._core.event_api import EventRecordsFilter

        latest_record = next(
            iter(
                self.instance.get_event_records(
                    event_records_filter=EventRecordsFilter(event_type=event_type),
                    limit=1,
                )
            ),
            None,
        )
        if latest_record is not None:
            return latest_record.storage_id
        return None

    @cached_method
    def _get_latest_materialization_or_observation_storage_ids_by_asset_partition(
        self, *, asset_key: AssetKey
    ) -> Mapping[AssetKeyPartitionKey, Optional[int]]:
        """Returns a mapping from asset partition to the latest storage id for that asset partition
        for all asset partitions associated with the given asset key.

        Note that for partitioned assets, an asset partition with a None partition key will be
        present in the mapping, representing the latest storage id for the asset as a whole.
        """
        asset_partition = AssetKeyPartitionKey(asset_key)
        latest_record = self._get_latest_materialization_or_observation_record(
            asset_partition=asset_partition
        )
        latest_storage_ids = {
            asset_partition: latest_record.storage_id if latest_record is not None else None
        }
        if self.asset_graph.is_partitioned(asset_key):
            latest_storage_ids.update(
                {
                    AssetKeyPartitionKey(asset_key, partition_key): storage_id
                    for partition_key, storage_id in self.instance.get_latest_storage_id_by_partition(
                        asset_key, event_type=self._event_type_for_key(asset_key)
                    ).items()
                }
            )
        return latest_storage_ids

    def get_latest_materialization_or_observation_storage_id(
        self, asset_partition: AssetKeyPartitionKey
    ) -> Optional[int]:
        """Returns the latest storage id for the given asset partition. If the asset has never been
        materialized, returns None.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
        """
        return self._get_latest_materialization_or_observation_storage_ids_by_asset_partition(
            asset_key=asset_partition.asset_key
        ).get(asset_partition)

    def asset_partition_has_materialization_or_observation(
        self,
        asset_partition: AssetKeyPartitionKey,
        after_cursor: Optional[int] = None,
    ) -> bool:
        """Returns True if there is a materialization record for the given asset partition after
        the specified cursor.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
            after_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                greater than this value will be considered.
        """
        return (self.get_latest_materialization_or_observation_storage_id(asset_partition) or 0) > (
            after_cursor or 0
        )

    def get_latest_materialization_or_observation_record(
        self,
        asset_partition: AssetKeyPartitionKey,
        after_cursor: Optional[int] = None,
        before_cursor: Optional[int] = None,
    ) -> Optional["EventLogRecord"]:
        """Returns the latest record for the given asset partition given the specified cursors.

        Args:
            asset_partition (AssetKeyPartitionKey): The asset partition to query.
            after_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                greater than this value will be considered.
            before_cursor (Optional[int]): Filter parameter such that only records with a storage_id
                less than this value will be considered.
        """
        check.param_invariant(
            not (after_cursor and before_cursor),
            "before_cursor",
            "Cannot set both before_cursor and after_cursor",
        )

        # first, do a quick check to eliminate the case where we know there is no record
        if not self.asset_partition_has_materialization_or_observation(
            asset_partition, after_cursor
        ):
            return None
        # then, if the before_cursor is after our latest record's storage id, we can just return
        # the latest record
        elif (before_cursor or 0) > (
            self.get_latest_materialization_or_observation_storage_id(asset_partition) or 0
        ):
            return self._get_latest_materialization_or_observation_record(
                asset_partition=asset_partition
            )
        # otherwise, do the explicit query
        return self._get_latest_materialization_or_observation_record(
            asset_partition=asset_partition, before_cursor=before_cursor
        )

    ####################
    # OBSERVATIONS
    ####################

    @cached_method
    def next_version_record(
        self,
        *,
        asset_key: AssetKey,
        after_cursor: Optional[int],
        data_version: Optional[DataVersion],
    ) -> Optional["EventLogRecord"]:
        from dagster._core.event_api import EventRecordsFilter

        for record in self.instance.get_event_records(
            EventRecordsFilter(
                event_type=DagsterEventType.ASSET_OBSERVATION,
                asset_key=asset_key,
                after_cursor=after_cursor,
            ),
            ascending=True,
        ):
            record_version = extract_data_version_from_entry(record.event_log_entry)
            if record_version is not None and record_version != data_version:
                return record

        # no records found with a new data version
        return None

    ####################
    # RUNS
    ####################

    @cached_method
    def _get_run_record_by_id(self, *, run_id: str) -> Optional[RunRecord]:
        return self.instance.get_run_record_by_id(run_id)

    def _get_run_by_id(self, run_id: str) -> Optional[DagsterRun]:
        run_record = self._get_run_record_by_id(run_id=run_id)
        if run_record is not None:
            return run_record.dagster_run
        return None

    def run_has_tag(self, run_id: str, tag_key: str, tag_value: str) -> bool:
        return cast(DagsterRun, self._get_run_by_id(run_id)).tags.get(tag_key) == tag_value

    @cached_method
    def _get_planned_materializations_for_run_from_events(
        self, *, run_id: str
    ) -> AbstractSet[AssetKey]:
        """Provides a fallback for fetching the planned materializations for a run from
        the ASSET_MATERIALIZATION_PLANNED events in the event log, in cases where this information
        is not available on the DagsterRun object.

        Args:
            run_id (str): The run id
        """
        materializations_planned = self.instance.get_records_for_run(
            run_id=run_id, of_type=DagsterEventType.ASSET_MATERIALIZATION_PLANNED
        ).records
        return set(cast(AssetKey, record.asset_key) for record in materializations_planned)

    def get_planned_materializations_for_run(self, run_id: str) -> AbstractSet[AssetKey]:
        """Returns the set of asset keys that are planned to be materialized by the run.

        Args:
            run_id (str): The run id
        """
        run = self._get_run_by_id(run_id=run_id)
        if run is None:
            return set()
        elif run.asset_selection:
            return run.asset_selection
        else:
            # must resort to querying the event log
            return self._get_planned_materializations_for_run_from_events(run_id=run_id)

    def is_asset_planned_for_run(
        self, run_id: str, asset: Union[AssetKey, AssetKeyPartitionKey]
    ) -> bool:
        """Returns True if the asset is planned to be materialized by the run."""
        run = self._get_run_by_id(run_id=run_id)
        if not run:
            return False

        if isinstance(asset, AssetKeyPartitionKey):
            asset_key = asset.asset_key
            if run.tags.get(PARTITION_NAME_TAG) != asset.partition_key:
                return False
        else:
            asset_key = asset

        return asset_key in self.get_planned_materializations_for_run(run_id=run_id)

    @cached_method
    def get_current_materializations_for_run(self, *, run_id: str) -> AbstractSet[AssetKey]:
        """Returns the set of asset keys that have been materialized by a given run.

        Args:
            run_id (str): The run id
        """
        materializations = self.instance.get_records_for_run(
            run_id=run_id,
            of_type=DagsterEventType.ASSET_MATERIALIZATION,
        ).records
        return set(cast(AssetKey, record.asset_key) for record in materializations)

    ####################
    # PARTITION COUNTS
    ####################

    def get_materialized_partition_counts(
        self, asset_key: AssetKey, after_cursor: Optional[int] = None
    ) -> Mapping[str, int]:
        """Returns a mapping from partition key to the number of times that partition has been
        materialized for a given asset.

        Args:
            asset_key (AssetKey): The asset key.
            after_cursor (Optional[int]): The cursor after which to look for materializations. If
                not provided, will look at all materializations.
        """
        if (
            after_cursor not in self._asset_partition_count_cache
            or asset_key not in self._asset_partition_count_cache[after_cursor]
        ):
            self._asset_partition_count_cache[after_cursor][
                asset_key
            ] = self.instance.get_materialization_count_by_partition(
                asset_keys=[asset_key], after_cursor=after_cursor
            )[
                asset_key
            ]
        return self._asset_partition_count_cache[after_cursor][asset_key]

    def get_materialized_partitions(
        self, asset_key: AssetKey, after_cursor: Optional[int] = None
    ) -> Iterable[str]:
        """Returns a list of the partitions that have been materialized for the given asset key.

        Args:
            asset_key (AssetKey): The asset key.
            after_cursor (Optional[int]): The cursor after which to look for materializations. If
                not provided, will look at all materializations.
        """
        return [
            partition_key
            for partition_key, count in self.get_materialized_partition_counts(
                asset_key, after_cursor
            ).items()
            if count > 0
        ]

    ####################
    # DYNAMIC PARTITIONS
    ####################

    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        """Returns a list of partitions for a partitions definition."""
        if partitions_def_name not in self._dynamic_partitions_cache:
            self._dynamic_partitions_cache[
                partitions_def_name
            ] = self.instance.get_dynamic_partitions(partitions_def_name)
        return self._dynamic_partitions_cache[partitions_def_name]

    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        return partition_key in self.get_dynamic_partitions(partitions_def_name)

    ####################
    # RECONCILIATION
    ####################

    def _asset_partitions_data_versions(
        self,
        asset_key: AssetKey,
        asset_partitions: Optional[Sequence[AssetKeyPartitionKey]],
        after_cursor: Optional[int] = None,
        before_cursor: Optional[int] = None,
    ) -> Mapping[AssetKeyPartitionKey, Optional[DataVersion]]:
        if not self.asset_graph.is_partitioned(asset_key):
            asset_partition = AssetKeyPartitionKey(asset_key)
            latest_record = self.get_latest_materialization_or_observation_record(
                asset_partition, after_cursor=after_cursor, before_cursor=before_cursor
            )
            return (
                {asset_partition: extract_data_version_from_entry(latest_record.event_log_entry)}
                if latest_record is not None
                else {}
            )
        else:
            query_result = self.instance._event_storage.get_latest_tags_by_partition(  # noqa
                asset_key,
                event_type=self._event_type_for_key(asset_key),
                tag_keys=[DATA_VERSION_TAG],
                after_cursor=after_cursor,
                before_cursor=before_cursor,
                asset_partitions=[
                    asset_partition.partition_key
                    for asset_partition in asset_partitions
                    if asset_partition.partition_key is not None
                ]
                if asset_partitions is not None
                else None,
            )
            return {
                AssetKeyPartitionKey(asset_key, partition_key): DataVersion(tags[DATA_VERSION_TAG])
                if tags.get(DATA_VERSION_TAG)
                else None
                for partition_key, tags in query_result.items()
            }

    def get_asset_partitions_updated_after_cursor(
        self,
        asset_key: AssetKey,
        asset_partitions: Optional[Sequence[AssetKeyPartitionKey]],
        after_cursor: Optional[int],
    ) -> Sequence[AssetKeyPartitionKey]:
        """Returns the set of asset partitions that have been updated after the given cursor.

        Args:
            asset_key (AssetKey): The asset key to check.
            asset_partitions (Optional[Sequence[AssetKeyPartitionKey]]): If supplied, will filter
                the set of checked partitions to the given partitions.
            after_cursor (Optional[int]): The cursor after which to look for updates.
        """
        if not self.asset_partition_has_materialization_or_observation(
            AssetKeyPartitionKey(asset_key), after_cursor=after_cursor
        ):
            return []
        # quick check that just compares latest storage ids
        updated_after_cursor = [
            asset_partition
            for asset_partition, latest_storage_id in self._get_latest_materialization_or_observation_storage_ids_by_asset_partition(
                asset_key=asset_key
            ).items()
            if (asset_partitions is None or asset_partition in asset_partitions)
            and (latest_storage_id or 0) > (after_cursor or 0)
        ]
        if not updated_after_cursor:
            return []
        # for regular assets, don't bother checking versions
        # in the future, we may remove this guard
        if after_cursor is None or not self.asset_graph.is_observable(asset_key):
            return updated_after_cursor

        # more expensive check to explicitly handle data versions
        latest_versions = self._asset_partitions_data_versions(
            asset_key, updated_after_cursor, after_cursor=after_cursor
        )
        previous_versions = self._asset_partitions_data_versions(
            asset_key, updated_after_cursor, before_cursor=after_cursor + 1
        )
        return [
            asset_partition
            for asset_partition, version in latest_versions.items()
            if previous_versions.get(asset_partition) != version
        ]

    @cached_method
    def is_reconciled(self, *, asset_partition: AssetKeyPartitionKey) -> bool:
        """Returns a boolean representing if the given `asset_partition` is currently reconciled.
        An asset (partition) is considered unreconciled if any of:
        - It has never been materialized
        - One of its parents has been updated more recently than it has
        - One of its parents is unreconciled.
        """
        # always treat source assets as reconciled
        if self.asset_graph.is_source(asset_partition.asset_key):
            return True
        elif not self.asset_partition_has_materialization_or_observation(asset_partition):
            return False

        time_or_dynamic_partitioned = isinstance(
            self.asset_graph.get_partitions_def(asset_partition.asset_key),
            (TimeWindowPartitionsDefinition, DynamicPartitionsDefinition),
        )

        parent_asset_partitions_by_key = defaultdict(list)

        for parent in self.asset_graph.get_parents_partitions(
            dynamic_partitions_store=self,
            current_time=self._evaluation_time,
            asset_key=asset_partition.asset_key,
            partition_key=asset_partition.partition_key,
        ).parent_partitions:
            parent_asset_partitions_by_key[parent.asset_key].append(parent)

        for parent_key, parent_asset_partitions in parent_asset_partitions_by_key.items():
            # ignore non-observable source parents
            if self.asset_graph.is_source(parent_key) and not self.asset_graph.is_observable(
                parent_key
            ):
                continue

            # when mapping from time or dynamic downstream to unpartitioned upstream, only check
            # for existence of upstream materialization, do not worry about timestamps
            if time_or_dynamic_partitioned and not self.asset_graph.is_partitioned(parent_key):
                return all(
                    self.asset_partition_has_materialization_or_observation(parent)
                    for parent in parent_asset_partitions
                )
            elif self.get_asset_partitions_updated_after_cursor(
                asset_key=parent_key,
                asset_partitions=parent_asset_partitions,
                after_cursor=self.get_latest_materialization_or_observation_storage_id(
                    asset_partition
                ),
            ):
                return False
            elif any(
                not self.is_reconciled(asset_partition=parent) for parent in parent_asset_partitions
            ):
                return False

        return True



from typing import Optional, Sequence

import dagster._check as check
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.host_representation import CodeLocation
from dagster._core.host_representation.external import ExternalJob
from dagster._core.host_representation.origin import ExternalJobOrigin


def external_job_from_location(
    code_location: CodeLocation,
    external_job_origin: ExternalJobOrigin,
    op_selection: Optional[Sequence[str]],
) -> ExternalJob:
    check.inst_param(code_location, "code_location", CodeLocation)
    check.inst_param(external_job_origin, "external_pipeline_origin", ExternalJobOrigin)

    repo_name = external_job_origin.external_repository_origin.repository_name
    job_name = external_job_origin.job_name

    check.invariant(
        code_location.has_repository(repo_name),
        f"Could not find repository {repo_name} in location {code_location.name}",
    )
    external_repo = code_location.get_repository(repo_name)

    pipeline_selector = JobSubsetSelector(
        location_name=code_location.name,
        repository_name=external_repo.name,
        job_name=job_name,
        op_selection=op_selection,
    )

    return code_location.get_external_job(pipeline_selector)



from .base import RunStorage as RunStorage
from .in_memory import InMemoryRunStorage as InMemoryRunStorage
from .schema import (
    DaemonHeartbeatsTable as DaemonHeartbeatsTable,
    InstanceInfo as InstanceInfo,
    RunStorageSqlMetadata as RunStorageSqlMetadata,
)
from .sql_run_storage import SqlRunStorage as SqlRunStorage
from .sqlite import SqliteRunStorage as SqliteRunStorage



from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from typing_extensions import TypeAlias, TypeVar

import dagster._check as check
from dagster._annotations import PublicAttr, public
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._serdes.serdes import (
    whitelist_for_serdes,
)
from dagster._utils import hash_collection

from .hook_definition import HookDefinition
from .input import FanInInputPointer, InputDefinition, InputMapping, InputPointer
from .output import OutputDefinition
from .utils import DEFAULT_OUTPUT, struct_to_string, validate_tags

if TYPE_CHECKING:
    from dagster._core.definitions.op_definition import OpDefinition

    from .asset_layer import AssetLayer
    from .composition import MappedInputPlaceholder
    from .graph_definition import GraphDefinition
    from .node_definition import NodeDefinition
    from .resource_requirement import ResourceRequirement

T_DependencyKey = TypeVar("T_DependencyKey", str, "NodeInvocation")
DependencyMapping: TypeAlias = Mapping[T_DependencyKey, Mapping[str, "IDependencyDefinition"]]


class NodeInvocation(
    NamedTuple(
        "Node",
        [
            ("name", PublicAttr[str]),
            ("alias", PublicAttr[Optional[str]]),
            ("tags", PublicAttr[Mapping[str, Any]]),
            ("hook_defs", PublicAttr[AbstractSet[HookDefinition]]),
            ("retry_policy", PublicAttr[Optional[RetryPolicy]]),
        ],
    )
):
    """Identifies an instance of a node in a graph dependency structure.

    Args:
        name (str): Name of the node of which this is an instance.
        alias (Optional[str]): Name specific to this instance of the node. Necessary when there are
            multiple instances of the same node.
        tags (Optional[Dict[str, Any]]): Optional tags values to extend or override those
            set on the node definition.
        hook_defs (Optional[AbstractSet[HookDefinition]]): A set of hook definitions applied to the
            node instance.

    Examples:
    In general, users should prefer not to construct this class directly or use the
    :py:class:`JobDefinition` API that requires instances of this class. Instead, use the
    :py:func:`@job <job>` API:

    .. code-block:: python

        from dagster import job

        @job
        def my_job():
            other_name = some_op.alias('other_name')
            some_graph(other_name(some_op))

    """

    def __new__(
        cls,
        name: str,
        alias: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
        retry_policy: Optional[RetryPolicy] = None,
    ):
        return super().__new__(
            cls,
            name=check.str_param(name, "name"),
            alias=check.opt_str_param(alias, "alias"),
            tags=check.opt_mapping_param(tags, "tags", value_type=str, key_type=str),
            hook_defs=check.opt_set_param(hook_defs, "hook_defs", of_type=HookDefinition),
            retry_policy=check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy),
        )

    # Needs to be hashable because this class is used as a key in dependencies dicts
    def __hash__(self) -> int:
        if not hasattr(self, "_hash"):
            self._hash = hash_collection(self)
        return self._hash


class Node(ABC):
    """Node invocation within a graph. Identified by its name inside the graph."""

    name: str
    definition: "NodeDefinition"
    graph_definition: "GraphDefinition"
    _additional_tags: Mapping[str, str]
    _hook_defs: AbstractSet[HookDefinition]
    _retry_policy: Optional[RetryPolicy]
    _inputs: Mapping[str, "NodeInput"]
    _outputs: Mapping[str, "NodeOutput"]

    def __init__(
        self,
        name: str,
        definition: "NodeDefinition",
        graph_definition: "GraphDefinition",
        tags: Optional[Mapping[str, str]] = None,
        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
        retry_policy: Optional[RetryPolicy] = None,
    ):
        from .graph_definition import GraphDefinition
        from .node_definition import NodeDefinition

        self.name = check.str_param(name, "name")
        self.definition = check.inst_param(definition, "definition", NodeDefinition)
        self.graph_definition = check.inst_param(
            graph_definition,
            "graph_definition",
            GraphDefinition,
        )
        self._additional_tags = validate_tags(tags)
        self._hook_defs = check.opt_set_param(hook_defs, "hook_defs", of_type=HookDefinition)
        self._retry_policy = check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy)

        self._inputs = {
            name: NodeInput(self, input_def)
            for name, input_def in self.definition.input_dict.items()
        }
        self._outputs = {
            name: NodeOutput(self, output_def)
            for name, output_def in self.definition.output_dict.items()
        }

    def inputs(self) -> Iterable["NodeInput"]:
        return self._inputs.values()

    def outputs(self) -> Iterable["NodeOutput"]:
        return self._outputs.values()

    def get_input(self, name: str) -> "NodeInput":
        check.str_param(name, "name")
        return self._inputs[name]

    def get_output(self, name: str) -> "NodeOutput":
        check.str_param(name, "name")
        return self._outputs[name]

    def has_input(self, name: str) -> bool:
        return self.definition.has_input(name)

    def input_def_named(self, name: str) -> InputDefinition:
        return self.definition.input_def_named(name)

    def has_output(self, name: str) -> bool:
        return self.definition.has_output(name)

    def output_def_named(self, name: str) -> OutputDefinition:
        return self.definition.output_def_named(name)

    @property
    def input_dict(self) -> Mapping[str, InputDefinition]:
        return self.definition.input_dict

    @property
    def output_dict(self) -> Mapping[str, OutputDefinition]:
        return self.definition.output_dict

    @property
    def tags(self) -> Mapping[str, str]:
        return {**self.definition.tags, **self._additional_tags}

    def container_maps_input(self, input_name: str) -> bool:
        return (
            self.graph_definition.input_mapping_for_pointer(InputPointer(self.name, input_name))
            is not None
        )

    def container_mapped_input(self, input_name: str) -> InputMapping:
        mapping = self.graph_definition.input_mapping_for_pointer(
            InputPointer(self.name, input_name)
        )
        if mapping is None:
            check.failed(
                f"container does not map input {input_name}, check container_maps_input first"
            )
        return mapping

    def container_maps_fan_in_input(self, input_name: str, fan_in_index: int) -> bool:
        return (
            self.graph_definition.input_mapping_for_pointer(
                FanInInputPointer(self.name, input_name, fan_in_index)
            )
            is not None
        )

    def container_mapped_fan_in_input(self, input_name: str, fan_in_index: int) -> InputMapping:
        mapping = self.graph_definition.input_mapping_for_pointer(
            FanInInputPointer(self.name, input_name, fan_in_index)
        )
        if mapping is None:
            check.failed(
                f"container does not map fan-in {input_name} idx {fan_in_index}, check "
                "container_maps_fan_in_input first"
            )

        return mapping

    @property
    def hook_defs(self) -> AbstractSet[HookDefinition]:
        return self._hook_defs

    @property
    def retry_policy(self) -> Optional[RetryPolicy]:
        return self._retry_policy

    @abstractmethod
    def describe_node(self) -> str:
        ...

    @abstractmethod
    def get_resource_requirements(
        self,
        outer_container: "GraphDefinition",
        parent_handle: Optional["NodeHandle"] = None,
        asset_layer: Optional["AssetLayer"] = None,
    ) -> Iterator["ResourceRequirement"]:
        ...


class GraphNode(Node):
    definition: "GraphDefinition"

    def __init__(
        self,
        name: str,
        definition: "GraphDefinition",
        graph_definition: "GraphDefinition",
        tags: Optional[Mapping[str, str]] = None,
        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
        retry_policy: Optional[RetryPolicy] = None,
    ):
        from .graph_definition import GraphDefinition

        check.inst_param(definition, "definition", GraphDefinition)
        super().__init__(name, definition, graph_definition, tags, hook_defs, retry_policy)

    def get_resource_requirements(
        self,
        outer_container: "GraphDefinition",
        parent_handle: Optional["NodeHandle"] = None,
        asset_layer: Optional["AssetLayer"] = None,
    ) -> Iterator["ResourceRequirement"]:
        cur_node_handle = NodeHandle(self.name, parent_handle)

        for node in self.definition.node_dict.values():
            yield from node.get_resource_requirements(
                asset_layer=asset_layer,
                outer_container=self.definition,
                parent_handle=cur_node_handle,
            )

    def describe_node(self) -> str:
        return f"graph '{self.name}'"


class OpNode(Node):
    definition: "OpDefinition"

    def __init__(
        self,
        name: str,
        definition: "OpDefinition",
        graph_definition: "GraphDefinition",
        tags: Optional[Mapping[str, str]] = None,
        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
        retry_policy: Optional[RetryPolicy] = None,
    ):
        from .op_definition import OpDefinition

        check.inst_param(definition, "definition", OpDefinition)
        super().__init__(name, definition, graph_definition, tags, hook_defs, retry_policy)

    def get_resource_requirements(
        self,
        outer_container: "GraphDefinition",
        parent_handle: Optional["NodeHandle"] = None,
        asset_layer: Optional["AssetLayer"] = None,
    ) -> Iterator["ResourceRequirement"]:
        from .resource_requirement import InputManagerRequirement

        cur_node_handle = NodeHandle(self.name, parent_handle)

        for requirement in self.definition.get_resource_requirements(
            (cur_node_handle, asset_layer)
        ):
            # If requirement is a root input manager requirement, but the corresponding node has an upstream output, then ignore the requirement.
            if (
                isinstance(requirement, InputManagerRequirement)
                and outer_container.dependency_structure.has_deps(
                    NodeInput(self, self.definition.input_def_named(requirement.input_name))
                )
                and requirement.root_input
            ):
                continue
            yield requirement
        for hook_def in self.hook_defs:
            yield from hook_def.get_resource_requirements(self.describe_node())

    def describe_node(self) -> str:
        return f"op '{self.name}'"


@whitelist_for_serdes(storage_name="SolidHandle")
class NodeHandle(NamedTuple("_NodeHandle", [("name", str), ("parent", Optional["NodeHandle"])])):
    """A structured object to identify nodes in the potentially recursive graph structure."""

    def __new__(cls, name: str, parent: Optional["NodeHandle"]):
        return super(NodeHandle, cls).__new__(
            cls,
            check.str_param(name, "name"),
            check.opt_inst_param(parent, "parent", NodeHandle),
        )

    def __str__(self):
        return self.to_string()

    @property
    def root(self):
        if self.parent:
            return self.parent.root
        else:
            return self

    @property
    def path(self) -> Sequence[str]:
        """Return a list representation of the handle.

        Inverse of NodeHandle.from_path.

        Returns:
            List[str]:
        """
        path: List[str] = []
        cur = self
        while cur:
            path.append(cur.name)
            cur = cur.parent
        path.reverse()
        return path

    def to_string(self) -> str:
        """Return a unique string representation of the handle.

        Inverse of NodeHandle.from_string.
        """
        return self.parent.to_string() + "." + self.name if self.parent else self.name

    def is_or_descends_from(self, handle: "NodeHandle") -> bool:
        """Check if the handle is or descends from another handle.

        Args:
            handle (NodeHandle): The handle to check against.

        Returns:
            bool:
        """
        check.inst_param(handle, "handle", NodeHandle)

        for idx in range(len(handle.path)):
            if idx >= len(self.path):
                return False
            if self.path[idx] != handle.path[idx]:
                return False
        return True

    def pop(self, ancestor: "NodeHandle") -> Optional["NodeHandle"]:
        """Return a copy of the handle with some of its ancestors pruned.

        Args:
            ancestor (NodeHandle): Handle to an ancestor of the current handle.

        Returns:
            NodeHandle:

        Example:
        .. code-block:: python

            handle = NodeHandle('baz', NodeHandle('bar', NodeHandle('foo', None)))
            ancestor = NodeHandle('bar', NodeHandle('foo', None))
            assert handle.pop(ancestor) == NodeHandle('baz', None)
        """
        check.inst_param(ancestor, "ancestor", NodeHandle)
        check.invariant(
            self.is_or_descends_from(ancestor),
            "Handle {handle} does not descend from {ancestor}".format(
                handle=self.to_string(), ancestor=ancestor.to_string()
            ),
        )

        return NodeHandle.from_path(self.path[len(ancestor.path) :])

    def with_ancestor(self, ancestor: Optional["NodeHandle"]) -> "NodeHandle":
        """Returns a copy of the handle with an ancestor grafted on.

        Args:
            ancestor (NodeHandle): Handle to the new ancestor.

        Returns:
            NodeHandle:

        Example:
        .. code-block:: python

            handle = NodeHandle('baz', NodeHandle('bar', NodeHandle('foo', None)))
            ancestor = NodeHandle('quux' None)
            assert handle.with_ancestor(ancestor) == NodeHandle(
                'baz', NodeHandle('bar', NodeHandle('foo', NodeHandle('quux', None)))
            )
        """
        check.opt_inst_param(ancestor, "ancestor", NodeHandle)

        return NodeHandle.from_path([*(ancestor.path if ancestor else []), *self.path])

    @staticmethod
    def from_path(path: Sequence[str]) -> "NodeHandle":
        check.sequence_param(path, "path", of_type=str)

        cur: Optional["NodeHandle"] = None
        _path = list(path)
        while len(_path) > 0:
            cur = NodeHandle(name=_path.pop(0), parent=cur)

        if cur is None:
            check.failed(f"Invalid handle path {path}")

        return cur

    @staticmethod
    def from_string(handle_str: str) -> "NodeHandle":
        check.str_param(handle_str, "handle_str")

        path = handle_str.split(".")
        return NodeHandle.from_path(path)

    @classmethod
    def from_dict(cls, dict_repr: Mapping[str, Any]) -> "NodeHandle":
        """This method makes it possible to load a potentially nested NodeHandle after a
        roundtrip through json.loads(json.dumps(NodeHandle._asdict())).
        """
        check.dict_param(dict_repr, "dict_repr", key_type=str)
        check.invariant(
            "name" in dict_repr, "Dict representation of NodeHandle must have a 'name' key"
        )
        check.invariant(
            "parent" in dict_repr, "Dict representation of NodeHandle must have a 'parent' key"
        )

        if isinstance(dict_repr["parent"], (list, tuple)):
            parent = NodeHandle.from_dict(
                {
                    "name": dict_repr["parent"][0],
                    "parent": dict_repr["parent"][1],
                }
            )
        else:
            parent = dict_repr["parent"]

        return NodeHandle(name=dict_repr["name"], parent=parent)


class NodeInputHandle(
    NamedTuple("_NodeInputHandle", [("node_handle", NodeHandle), ("input_name", str)])
):
    """A structured object to uniquely identify inputs in the potentially recursive graph structure.
    """


class NodeOutputHandle(
    NamedTuple("_NodeOutputHandle", [("node_handle", NodeHandle), ("output_name", str)])
):
    """A structured object to uniquely identify outputs in the potentially recursive graph structure.
    """


class NodeInput(NamedTuple("_NodeInput", [("node", Node), ("input_def", InputDefinition)])):
    def __new__(cls, node: Node, input_def: InputDefinition):
        return super(NodeInput, cls).__new__(
            cls,
            check.inst_param(node, "node", Node),
            check.inst_param(input_def, "input_def", InputDefinition),
        )

    def _inner_str(self) -> str:
        return struct_to_string(
            "NodeInput",
            node_name=self.node.name,
            input_name=self.input_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self):
        return hash((self.node.name, self.input_def.name))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, NodeInput)
            and self.node.name == other.node.name
            and self.input_def.name == other.input_def.name
        )

    @property
    def node_name(self) -> str:
        return self.node.name

    @property
    def input_name(self) -> str:
        return self.input_def.name


class NodeOutput(NamedTuple("_NodeOutput", [("node", Node), ("output_def", OutputDefinition)])):
    def __new__(cls, node: Node, output_def: OutputDefinition):
        return super(NodeOutput, cls).__new__(
            cls,
            check.inst_param(node, "node", Node),
            check.inst_param(output_def, "output_def", OutputDefinition),
        )

    def _inner_str(self) -> str:
        return struct_to_string(
            "NodeOutput",
            node_name=self.node.name,
            output_name=self.output_def.name,
        )

    def __str__(self):
        return self._inner_str()

    def __repr__(self):
        return self._inner_str()

    def __hash__(self) -> int:
        return hash((self.node.name, self.output_def.name))

    def __eq__(self, other: Any) -> bool:
        return self.node.name == other.node.name and self.output_def.name == other.output_def.name

    def describe(self) -> str:
        return f"{self.node_name}:{self.output_def.name}"

    @property
    def node_name(self) -> str:
        return self.node.name

    @property
    def is_dynamic(self) -> bool:
        return self.output_def.is_dynamic

    @property
    def output_name(self) -> str:
        return self.output_def.name


class DependencyType(Enum):
    DIRECT = "DIRECT"
    FAN_IN = "FAN_IN"
    DYNAMIC_COLLECT = "DYNAMIC_COLLECT"


class IDependencyDefinition(ABC):
    @abstractmethod
    def get_node_dependencies(self) -> Sequence["DependencyDefinition"]:
        pass

    @abstractmethod
    def is_fan_in(self) -> bool:
        """The result passed to the corresponding input will be a List made from different node outputs.
        """


class DependencyDefinition(
    NamedTuple(
        "_DependencyDefinition", [("node", str), ("output", str), ("description", Optional[str])]
    ),
    IDependencyDefinition,
):
    """Represents an edge in the DAG of nodes (ops or graphs) forming a job.

    This object is used at the leaves of a dictionary structure that represents the complete
    dependency structure of a job whose keys represent the dependent node and dependent
    input, so this object only contains information about the dependee.

    Concretely, if the input named 'input' of op_b depends on the output named 'result' of
    op_a, and the output named 'other_result' of graph_a, the structure will look as follows:

    .. code-block:: python

        dependency_structure = {
            'my_downstream_op': {
                'input': DependencyDefinition('my_upstream_op', 'result')
            }
            'my_downstream_op': {
                'input': DependencyDefinition('my_upstream_graph', 'result')
            }
        }

    In general, users should prefer not to construct this class directly or use the
    :py:class:`JobDefinition` API that requires instances of this class. Instead, use the
    :py:func:`@job <job>` API:

    .. code-block:: python

        @job
        def the_job():
            node_b(node_a())


    Args:
        node (str): The name of the node (op or graph) that is depended on, that is, from which the value
            passed between the two nodes originates.
        output (Optional[str]): The name of the output that is depended on. (default: "result")
        description (Optional[str]): Human-readable description of this dependency.
    """

    def __new__(
        cls,
        node: str,
        output: str = DEFAULT_OUTPUT,
        description: Optional[str] = None,
    ):
        return super(DependencyDefinition, cls).__new__(
            cls,
            check.str_param(node, "node"),
            check.str_param(output, "output"),
            check.opt_str_param(description, "description"),
        )

    def get_node_dependencies(self) -> Sequence["DependencyDefinition"]:
        return [self]

    def is_fan_in(self) -> bool:
        return False

    def get_op_dependencies(self) -> Sequence["DependencyDefinition"]:
        return [self]


class MultiDependencyDefinition(
    NamedTuple(
        "_MultiDependencyDefinition",
        [
            (
                "dependencies",
                PublicAttr[Sequence[Union[DependencyDefinition, Type["MappedInputPlaceholder"]]]],
            )
        ],
    ),
    IDependencyDefinition,
):
    """Represents a fan-in edge in the DAG of op instances forming a job.

    This object is used only when an input of type ``List[T]`` is assembled by fanning-in multiple
    upstream outputs of type ``T``.

    This object is used at the leaves of a dictionary structure that represents the complete
    dependency structure of a job whose keys represent the dependent ops or graphs and dependent
    input, so this object only contains information about the dependee.

    Concretely, if the input named 'input' of op_c depends on the outputs named 'result' of
    op_a and op_b, this structure will look as follows:

    .. code-block:: python

        dependency_structure = {
            'op_c': {
                'input': MultiDependencyDefinition(
                    [
                        DependencyDefinition('op_a', 'result'),
                        DependencyDefinition('op_b', 'result')
                    ]
                )
            }
        }

    In general, users should prefer not to construct this class directly or use the
    :py:class:`JobDefinition` API that requires instances of this class. Instead, use the
    :py:func:`@job <job>` API:

    .. code-block:: python

        @job
        def the_job():
            op_c(op_a(), op_b())

    Args:
        dependencies (List[Union[DependencyDefinition, Type[MappedInputPlaceHolder]]]): List of
            upstream dependencies fanned in to this input.
    """

    def __new__(
        cls,
        dependencies: Sequence[Union[DependencyDefinition, Type["MappedInputPlaceholder"]]],
    ):
        from .composition import MappedInputPlaceholder

        deps = check.sequence_param(dependencies, "dependencies")
        seen = {}
        for dep in deps:
            if isinstance(dep, DependencyDefinition):
                key = dep.node + ":" + dep.output
                if key in seen:
                    raise DagsterInvalidDefinitionError(
                        f'Duplicate dependencies on node "{dep.node}" output "{dep.output}" '
                        "used in the same MultiDependencyDefinition."
                    )
                seen[key] = True
            elif dep is MappedInputPlaceholder:
                pass
            else:
                check.failed(f"Unexpected dependencies entry {dep}")

        return super(MultiDependencyDefinition, cls).__new__(cls, deps)

    @public
    def get_node_dependencies(self) -> Sequence[DependencyDefinition]:
        return [dep for dep in self.dependencies if isinstance(dep, DependencyDefinition)]

    @public
    def is_fan_in(self) -> bool:
        return True

    @public
    def get_dependencies_and_mappings(
        self,
    ) -> Sequence[Union[DependencyDefinition, Type["MappedInputPlaceholder"]]]:
        return self.dependencies


class DynamicCollectDependencyDefinition(
    NamedTuple("_DynamicCollectDependencyDefinition", [("node_name", str), ("output_name", str)]),
    IDependencyDefinition,
):
    def get_node_dependencies(self) -> Sequence[DependencyDefinition]:
        return [DependencyDefinition(self.node_name, self.output_name)]

    def is_fan_in(self) -> bool:
        return True


DepTypeAndOutputs: TypeAlias = Tuple[
    DependencyType,
    Union[NodeOutput, List[Union[NodeOutput, Type["MappedInputPlaceholder"]]]],
]

InputToOutputMap: TypeAlias = Dict[NodeInput, DepTypeAndOutputs]


def _create_handle_dict(
    node_dict: Mapping[str, Node],
    dep_dict: DependencyMapping[str],
) -> InputToOutputMap:
    from .composition import MappedInputPlaceholder

    check.mapping_param(node_dict, "node_dict", key_type=str, value_type=Node)
    check.two_dim_mapping_param(dep_dict, "dep_dict", value_type=IDependencyDefinition)

    handle_dict: InputToOutputMap = {}

    for node_name, input_dict in dep_dict.items():
        from_node = node_dict[node_name]
        for input_name, dep_def in input_dict.items():
            if isinstance(dep_def, MultiDependencyDefinition):
                handles: List[Union[NodeOutput, Type[MappedInputPlaceholder]]] = []
                for inner_dep in dep_def.get_dependencies_and_mappings():
                    if isinstance(inner_dep, DependencyDefinition):
                        handles.append(node_dict[inner_dep.node].get_output(inner_dep.output))
                    elif inner_dep is MappedInputPlaceholder:
                        handles.append(inner_dep)
                    else:
                        check.failed(
                            "Unexpected MultiDependencyDefinition dependencies type {}".format(
                                inner_dep
                            )
                        )

                handle_dict[from_node.get_input(input_name)] = (DependencyType.FAN_IN, handles)

            elif isinstance(dep_def, DependencyDefinition):
                handle_dict[from_node.get_input(input_name)] = (
                    DependencyType.DIRECT,
                    node_dict[dep_def.node].get_output(dep_def.output),
                )
            elif isinstance(dep_def, DynamicCollectDependencyDefinition):
                handle_dict[from_node.get_input(input_name)] = (
                    DependencyType.DYNAMIC_COLLECT,
                    node_dict[dep_def.node_name].get_output(dep_def.output_name),
                )

            else:
                check.failed(f"Unknown dependency type {dep_def}")

    return handle_dict


class DependencyStructure:
    @staticmethod
    def from_definitions(
        nodes: Mapping[str, Node], dep_dict: DependencyMapping[str]
    ) -> "DependencyStructure":
        return DependencyStructure(list(dep_dict.keys()), _create_handle_dict(nodes, dep_dict))

    _node_input_index: DefaultDict[str, Dict[NodeInput, List[NodeOutput]]]
    _node_output_index: Dict[str, DefaultDict[NodeOutput, List[NodeInput]]]
    _dynamic_fan_out_index: Dict[str, NodeOutput]
    _collect_index: Dict[str, Set[NodeOutput]]

    def __init__(self, node_names: Sequence[str], input_to_output_map: InputToOutputMap):
        self._node_names = node_names
        self._input_to_output_map = input_to_output_map

        # Building up a couple indexes here so that one can look up all the upstream output handles
        # or downstream input handles in O(1). Without this, this can become O(N^2) where N is node
        # count during the GraphQL query in particular

        # node_name => input_handle => list[output_handle]
        self._node_input_index = defaultdict(dict)

        # node_name => output_handle => list[input_handle]
        self._node_output_index = defaultdict(lambda: defaultdict(list))

        # node_name => dynamic output_handle that this node will dupe for
        self._dynamic_fan_out_index = {}

        # node_name => set of dynamic output_handle this collects over
        self._collect_index = defaultdict(set)

        for node_input, (dep_type, node_output_or_list) in self._input_to_output_map.items():
            if dep_type == DependencyType.FAN_IN:
                node_output_list: List[NodeOutput] = []
                for node_output in node_output_or_list:
                    if not isinstance(node_output, NodeOutput):
                        continue

                    if node_output.is_dynamic:
                        raise DagsterInvalidDefinitionError(
                            "Currently, items in a fan-in dependency cannot be downstream of"
                            " dynamic outputs. Problematic dependency on dynamic output"
                            f' "{node_output.describe()}".'
                        )
                    if self._dynamic_fan_out_index.get(node_output.node_name):
                        raise DagsterInvalidDefinitionError(
                            "Currently, items in a fan-in dependency cannot be downstream of"
                            " dynamic outputs. Problematic dependency on output"
                            f' "{node_output.describe()}", downstream of'
                            f' "{self._dynamic_fan_out_index[node_output.node_name].describe()}".'
                        )

                    node_output_list.append(node_output)
            elif dep_type == DependencyType.DIRECT:
                node_output = cast(NodeOutput, node_output_or_list)

                if node_output.is_dynamic:
                    self._validate_and_set_fan_out(node_input, node_output)

                if self._dynamic_fan_out_index.get(node_output.node_name):
                    self._validate_and_set_fan_out(
                        node_input, self._dynamic_fan_out_index[node_output.node_name]
                    )

                node_output_list = [node_output]
            elif dep_type == DependencyType.DYNAMIC_COLLECT:
                node_output = cast(NodeOutput, node_output_or_list)

                if node_output.is_dynamic:
                    self._validate_and_set_collect(node_input, node_output)

                elif self._dynamic_fan_out_index.get(node_output.node_name):
                    self._validate_and_set_collect(
                        node_input,
                        self._dynamic_fan_out_index[node_output.node_name],
                    )
                else:
                    check.failed(
                        f"Unexpected dynamic fan in dep created {node_output} -> {node_input}"
                    )

                node_output_list = [node_output]
            else:
                check.failed(f"Unexpected dep type {dep_type}")

            self._node_input_index[node_input.node.name][node_input] = node_output_list
            for node_output in node_output_list:
                self._node_output_index[node_output.node.name][node_output].append(node_input)

    def _validate_and_set_fan_out(self, node_input: NodeInput, node_output: NodeOutput) -> None:
        """Helper function for populating _dynamic_fan_out_index."""
        if not node_input.node.definition.input_supports_dynamic_output_dep(node_input.input_name):
            raise DagsterInvalidDefinitionError(
                f"{node_input.node.describe_node()} cannot be downstream of dynamic output"
                f' "{node_output.describe()}" since input "{node_input.input_name}" maps to a'
                " node that is already downstream of another dynamic output. Nodes cannot be"
                " downstream of more than one dynamic output"
            )

        if self._collect_index.get(node_input.node_name):
            raise DagsterInvalidDefinitionError(
                f"{node_input.node.describe_node()} cannot be both downstream of dynamic output "
                f"{node_output.describe()} and collect over dynamic output "
                f"{list(self._collect_index[node_input.node_name])[0].describe()}."
            )

        if self._dynamic_fan_out_index.get(node_input.node_name) is None:
            self._dynamic_fan_out_index[node_input.node_name] = node_output
            return

        if self._dynamic_fan_out_index[node_input.node_name] != node_output:
            raise DagsterInvalidDefinitionError(
                f"{node_input.node.describe_node()} cannot be downstream of more than one dynamic"
                f' output. It is downstream of both "{node_output.describe()}" and'
                f' "{self._dynamic_fan_out_index[node_input.node_name].describe()}"'
            )

    def _validate_and_set_collect(
        self,
        node_input: NodeInput,
        node_output: NodeOutput,
    ) -> None:
        if self._dynamic_fan_out_index.get(node_input.node_name):
            raise DagsterInvalidDefinitionError(
                f"{node_input.node.describe_node()} cannot both collect over dynamic output "
                f"{node_output.describe()} and be downstream of the dynamic output "
                f"{self._dynamic_fan_out_index[node_input.node_name].describe()}."
            )

        self._collect_index[node_input.node_name].add(node_output)

        # if the output is already fanned out
        if self._dynamic_fan_out_index.get(node_output.node_name):
            raise DagsterInvalidDefinitionError(
                f"{node_input.node.describe_node()} cannot be downstream of more than one dynamic"
                f' output. It is downstream of both "{node_output.describe()}" and'
                f' "{self._dynamic_fan_out_index[node_output.node_name].describe()}"'
            )

    def all_upstream_outputs_from_node(self, node_name: str) -> Sequence[NodeOutput]:
        check.str_param(node_name, "node_name")

        # flatten out all outputs that feed into the inputs of this node
        return [
            output_handle
            for output_handle_list in self._node_input_index[node_name].values()
            for output_handle in output_handle_list
        ]

    def input_to_upstream_outputs_for_node(
        self, node_name: str
    ) -> Mapping[NodeInput, Sequence[NodeOutput]]:
        """Returns a Dict[NodeInput, List[NodeOutput]] that encodes
        where all the the inputs are sourced from upstream. Usually the
        List[NodeOutput] will be a list of one, except for the
        multi-dependency case.
        """
        check.str_param(node_name, "node_name")
        return self._node_input_index[node_name]

    def output_to_downstream_inputs_for_node(
        self, node_name: str
    ) -> Mapping[NodeOutput, Sequence[NodeInput]]:
        """Returns a Dict[NodeOutput, List[NodeInput]] that
        represents all the downstream inputs for each output in the
        dictionary.
        """
        check.str_param(node_name, "node_name")
        return self._node_output_index[node_name]

    def has_direct_dep(self, node_input: NodeInput) -> bool:
        check.inst_param(node_input, "node_input", NodeInput)
        if node_input not in self._input_to_output_map:
            return False
        dep_type, _ = self._input_to_output_map[node_input]
        return dep_type == DependencyType.DIRECT

    def get_direct_dep(self, node_input: NodeInput) -> NodeOutput:
        check.inst_param(node_input, "node_input", NodeInput)
        dep_type, dep = self._input_to_output_map[node_input]
        check.invariant(
            dep_type == DependencyType.DIRECT,
            f"Cannot call get_direct_dep when dep is not singular, got {dep_type}",
        )
        return cast(NodeOutput, dep)

    def has_fan_in_deps(self, node_input: NodeInput) -> bool:
        check.inst_param(node_input, "node_input", NodeInput)
        if node_input not in self._input_to_output_map:
            return False
        dep_type, _ = self._input_to_output_map[node_input]
        return dep_type == DependencyType.FAN_IN

    def get_fan_in_deps(
        self, node_input: NodeInput
    ) -> Sequence[Union[NodeOutput, Type["MappedInputPlaceholder"]]]:
        check.inst_param(node_input, "node_input", NodeInput)
        dep_type, deps = self._input_to_output_map[node_input]
        check.invariant(
            dep_type == DependencyType.FAN_IN,
            f"Cannot call get_multi_dep when dep is not fan in, got {dep_type}",
        )
        return cast(List[Union[NodeOutput, Type["MappedInputPlaceholder"]]], deps)

    def has_dynamic_fan_in_dep(self, node_input: NodeInput) -> bool:
        check.inst_param(node_input, "node_input", NodeInput)
        if node_input not in self._input_to_output_map:
            return False
        dep_type, _ = self._input_to_output_map[node_input]
        return dep_type == DependencyType.DYNAMIC_COLLECT

    def get_dynamic_fan_in_dep(self, node_input: NodeInput) -> NodeOutput:
        check.inst_param(node_input, "node_input", NodeInput)
        dep_type, dep = self._input_to_output_map[node_input]
        check.invariant(
            dep_type == DependencyType.DYNAMIC_COLLECT,
            f"Cannot call get_dynamic_fan_in_dep when dep is not, got {dep_type}",
        )
        return cast(NodeOutput, dep)

    def has_deps(self, node_input: NodeInput) -> bool:
        check.inst_param(node_input, "node_input", NodeInput)
        return node_input in self._input_to_output_map

    def get_deps_list(self, node_input: NodeInput) -> Sequence[NodeOutput]:
        check.inst_param(node_input, "node_input", NodeInput)
        check.invariant(self.has_deps(node_input))
        dep_type, handle_or_list = self._input_to_output_map[node_input]
        if dep_type == DependencyType.DIRECT:
            return [cast(NodeOutput, handle_or_list)]
        elif dep_type == DependencyType.DYNAMIC_COLLECT:
            return [cast(NodeOutput, handle_or_list)]
        elif dep_type == DependencyType.FAN_IN:
            return [handle for handle in handle_or_list if isinstance(handle, NodeOutput)]
        else:
            check.failed(f"Unexpected dep type {dep_type}")

    def inputs(self) -> Sequence[NodeInput]:
        return list(self._input_to_output_map.keys())

    def get_upstream_dynamic_output_for_node(self, node_name: str) -> Optional[NodeOutput]:
        return self._dynamic_fan_out_index.get(node_name)

    def get_dependency_type(self, node_input: NodeInput) -> Optional[DependencyType]:
        result = self._input_to_output_map.get(node_input)
        if result is None:
            return None
        dep_type, _ = result
        return dep_type

    def is_dynamic_mapped(self, node_name: str) -> bool:
        return node_name in self._dynamic_fan_out_index

    def has_dynamic_downstreams(self, node_name: str) -> bool:
        for node_output in self._dynamic_fan_out_index.values():
            if node_output.node_name == node_name:
                return True

        return False



"""Tools for accessing core Dagster APIs over a GRPC mechanism.

GRPC is intended to be used in all cases where host processes communicate with user processes, both
locally (over UDS on MacOS and Unix, and over a local port on Windows) and when communicating with
remote Dagster user proceses (e.g., containers).

The GRPC layer is not intended to supplant the dagster-graphql layer, which should still be used to
drive web frontends like dagit.
"""

from .client import (
    DagsterGrpcClient as DagsterGrpcClient,
    client_heartbeat_thread as client_heartbeat_thread,
    ephemeral_grpc_api_client as ephemeral_grpc_api_client,
)
from .impl import core_execute_run as core_execute_run
from .server import (
    DagsterGrpcServer as DagsterGrpcServer,
    GrpcServerProcess as GrpcServerProcess,
)
from .types import (
    CanCancelExecutionRequest as CanCancelExecutionRequest,
    CanCancelExecutionResult as CanCancelExecutionResult,
    CancelExecutionRequest as CancelExecutionRequest,
    CancelExecutionResult as CancelExecutionResult,
    ExecuteExternalJobArgs as ExecuteExternalJobArgs,
    ExecuteRunArgs as ExecuteRunArgs,
    ExecuteStepArgs as ExecuteStepArgs,
    ExecutionPlanSnapshotArgs as ExecutionPlanSnapshotArgs,
    ExternalJobArgs as ExternalJobArgs,
    ExternalScheduleExecutionArgs as ExternalScheduleExecutionArgs,
    GetCurrentImageResult as GetCurrentImageResult,
    JobSubsetSnapshotArgs as JobSubsetSnapshotArgs,
    ListRepositoriesInput as ListRepositoriesInput,
    ListRepositoriesResponse as ListRepositoriesResponse,
    LoadableRepositorySymbol as LoadableRepositorySymbol,
    NotebookPathArgs as NotebookPathArgs,
    PartitionArgs as PartitionArgs,
    PartitionNamesArgs as PartitionNamesArgs,
    PartitionSetExecutionParamArgs as PartitionSetExecutionParamArgs,
    ResumeRunArgs as ResumeRunArgs,
    SensorExecutionArgs as SensorExecutionArgs,
    ShutdownServerResult as ShutdownServerResult,
    StartRunResult as StartRunResult,
)
from .utils import get_loadable_targets as get_loadable_targets



from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.sensor_definition import SensorExecutionData
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.host_representation.external_data import ExternalSensorExecutionErrorData
from dagster._core.host_representation.handle import RepositoryHandle
from dagster._grpc.client import DEFAULT_GRPC_TIMEOUT
from dagster._grpc.types import SensorExecutionArgs
from dagster._serdes import deserialize_value

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance
    from dagster._grpc.client import DagsterGrpcClient


def sync_get_external_sensor_execution_data_ephemeral_grpc(
    instance: "DagsterInstance",
    repository_handle: RepositoryHandle,
    sensor_name: str,
    last_completion_time: Optional[float],
    last_run_key: Optional[str],
    cursor: Optional[str],
    timeout: Optional[int] = DEFAULT_GRPC_TIMEOUT,
) -> SensorExecutionData:
    from dagster._grpc.client import ephemeral_grpc_api_client

    origin = repository_handle.get_external_origin()
    with ephemeral_grpc_api_client(
        origin.code_location_origin.loadable_target_origin
    ) as api_client:
        return sync_get_external_sensor_execution_data_grpc(
            api_client,
            instance,
            repository_handle,
            sensor_name,
            last_completion_time,
            last_run_key,
            cursor,
            timeout=timeout,
        )


def sync_get_external_sensor_execution_data_grpc(
    api_client: "DagsterGrpcClient",
    instance: "DagsterInstance",
    repository_handle: RepositoryHandle,
    sensor_name: str,
    last_completion_time: Optional[float],
    last_run_key: Optional[str],
    cursor: Optional[str],
    timeout: Optional[int] = DEFAULT_GRPC_TIMEOUT,
) -> SensorExecutionData:
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(sensor_name, "sensor_name")
    check.opt_float_param(last_completion_time, "last_completion_time")
    check.opt_str_param(last_run_key, "last_run_key")
    check.opt_str_param(cursor, "cursor")

    origin = repository_handle.get_external_origin()

    result = deserialize_value(
        api_client.external_sensor_execution(
            sensor_execution_args=SensorExecutionArgs(
                repository_origin=origin,
                instance_ref=instance.get_ref(),
                sensor_name=sensor_name,
                last_completion_time=last_completion_time,
                last_run_key=last_run_key,
                cursor=cursor,
            ),
            timeout=timeout,
        ),
        (SensorExecutionData, ExternalSensorExecutionErrorData),
    )

    if isinstance(result, ExternalSensorExecutionErrorData):
        raise DagsterUserCodeProcessError.from_error_info(result.error)

    return result



from typing import (
    Any,
)


class IAttachDifferentObjectToOpContext:
    """Utility interface which adjusts the type of the bare object attached to the execution context
    by a resource. Useful when migrating function-style resources to class-style ConfigurableResources.

    For example, the following function-style resource is consumed by an asset:

    .. code-block:: python

        @resource(config_schema={"inner_string": str})
        def my_string_resource(context) -> str:
            return context.resource_config["inner_string"]

        @asset(required_resource_keys={"my_string"})
        def my_unconverted_asset(context: AssetExecutionContext) -> str:
            return context.resources.my_string

    Adapted to a class-style resource, we can ensure our new ConfigurableResource is compatible
    with the asset by implementing this interface:

    .. code-block:: python

        class MyStringResource(ConfigurableResource, IAttachDifferentObjectToOpContext):
            inner_string: str

            def get_object_to_set_on_execution_context(self) -> str:
                return self.inner_string

        @asset(required_resource_keys={"my_string"})
        def my_unconverted_asset(context: AssetExecutionContext) -> str:
            return context.resources.my_string

        @asset
        def my_converted_asset(my_string: MyStringResource) -> str:
            return my_string.inner_string
    """

    def get_object_to_set_on_execution_context(self) -> Any:
        """Override to return the object to be attached to the execution context by this resource.
        """
        raise NotImplementedError()



from typing import (
    Callable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.errors import DagsterInvariantViolationError

from .valid_definitions import T_RepositoryLevelDefinition


class CacheingDefinitionIndex(Generic[T_RepositoryLevelDefinition]):
    def __init__(
        self,
        definition_class: Type[T_RepositoryLevelDefinition],
        definition_class_name: str,
        definition_kind: str,
        definitions: Mapping[
            str, Union[T_RepositoryLevelDefinition, Callable[[], T_RepositoryLevelDefinition]]
        ],
        validation_fn: Callable[[T_RepositoryLevelDefinition], T_RepositoryLevelDefinition],
        lazy_definitions_fn: Optional[Callable[[], Sequence[T_RepositoryLevelDefinition]]] = None,
    ):
        """Args:
        definitions: A dictionary of definition names to definitions or functions that load
            definitions.
        lazy_definitions_fn: A function for loading a list of definitions whose names are not
            even known until loaded.

        """
        for key, definition in definitions.items():
            check.invariant(
                isinstance(definition, definition_class) or callable(definition),
                "Bad definition for {definition_kind} {key}: must be {definition_class_name} or "
                "callable, got {type_}".format(
                    definition_kind=definition_kind,
                    key=key,
                    definition_class_name=definition_class_name,
                    type_=type(definition),
                ),
            )

        self._definition_class: Type[T_RepositoryLevelDefinition] = definition_class
        self._definition_class_name = definition_class_name
        self._definition_kind = definition_kind
        self._validation_fn: Callable[
            [T_RepositoryLevelDefinition], T_RepositoryLevelDefinition
        ] = validation_fn

        self._definitions: Mapping[
            str, Union[T_RepositoryLevelDefinition, Callable[[], T_RepositoryLevelDefinition]]
        ] = definitions
        self._definition_cache: Dict[str, T_RepositoryLevelDefinition] = {}
        self._definition_names: Optional[Sequence[str]] = None

        self._lazy_definitions_fn: Callable[
            [], Sequence[T_RepositoryLevelDefinition]
        ] = lazy_definitions_fn or (lambda: [])
        self._lazy_definitions: Optional[Sequence[T_RepositoryLevelDefinition]] = None

        self._all_definitions: Optional[Sequence[T_RepositoryLevelDefinition]] = None

    def _get_lazy_definitions(self) -> Sequence[T_RepositoryLevelDefinition]:
        if self._lazy_definitions is None:
            self._lazy_definitions = self._lazy_definitions_fn()
            for definition in self._lazy_definitions:
                self._validate_and_cache_definition(definition, definition.name)

        return self._lazy_definitions

    def get_definition_names(self) -> Sequence[str]:
        if self._definition_names:
            return self._definition_names

        lazy_names = []
        for definition in self._get_lazy_definitions():
            strict_definition = self._definitions.get(definition.name)
            if strict_definition:
                check.invariant(
                    strict_definition == definition,
                    f"Duplicate definition found for {definition.name}",
                )
            else:
                lazy_names.append(definition.name)

        self._definition_names = list(self._definitions.keys()) + lazy_names
        return self._definition_names

    def has_definition(self, definition_name: str) -> bool:
        check.str_param(definition_name, "definition_name")

        return definition_name in self.get_definition_names()

    def get_all_definitions(self) -> Sequence[T_RepositoryLevelDefinition]:
        if self._all_definitions is not None:
            return self._all_definitions

        self._all_definitions = list(
            sorted(
                map(self.get_definition, self.get_definition_names()),
                key=lambda definition: definition.name,
            )
        )
        return self._all_definitions

    def get_definition(self, definition_name: str) -> T_RepositoryLevelDefinition:
        check.str_param(definition_name, "definition_name")

        if not self.has_definition(definition_name):
            raise DagsterInvariantViolationError(
                "Could not find {definition_kind} '{definition_name}'. Found: "
                "{found_names}.".format(
                    definition_kind=self._definition_kind,
                    definition_name=definition_name,
                    found_names=", ".join(
                        [f"'{found_name}'" for found_name in self.get_definition_names()]
                    ),
                )
            )

        if definition_name in self._definition_cache:
            return self._definition_cache[definition_name]

        definition_source = self._definitions[definition_name]

        if isinstance(definition_source, self._definition_class):
            self._definition_cache[definition_name] = self._validation_fn(definition_source)
            return definition_source
        else:
            definition = cast(Callable, definition_source)()
            self._validate_and_cache_definition(definition, definition_name)
            return definition

    def _validate_and_cache_definition(
        self, definition: T_RepositoryLevelDefinition, definition_dict_key: str
    ):
        check.invariant(
            isinstance(definition, self._definition_class),
            "Bad constructor for {definition_kind} {definition_name}: must return "
            "{definition_class_name}, got value of type {type_}".format(
                definition_kind=self._definition_kind,
                definition_name=definition_dict_key,
                definition_class_name=self._definition_class_name,
                type_=type(definition),
            ),
        )
        check.invariant(
            definition.name == definition_dict_key,
            "Bad constructor for {definition_kind} '{definition_name}': name in "
            "{definition_class_name} does not match: got '{definition_def_name}'".format(
                definition_kind=self._definition_kind,
                definition_name=definition_dict_key,
                definition_class_name=self._definition_class_name,
                definition_def_name=definition.name,
            ),
        )
        self._definition_cache[definition_dict_key] = self._validation_fn(definition)
